import logging

_mapped_colors = {}
_remaining_colors = [
    "#66b54b",
    "#a65ecb",
    "#a5b13f",
    "#6976d6",
    "#cf9834",
    "#c5522e",
    "#c8489b",
    "#91b99d",
    "#d44368",
    "#63b8cc",
    "#56bc84",
    "#5a8dc5",
    "#db8566",
    "#8C85AF",
    "#b09e65",
    "#823570",
    "#48786d",
    "#d28cd7",
    "#cbd6ad",
    "#d596be",
    "#866a92",
    "#d883a0",
    "#afa6ce",
    "#7d5229",
    "#caa59a",
    "#a1607b",
    "#5e77da",
    "#97b942",
    "#caa43f",
    "#d8917e",
    "#74d17c",
    "#a6de9a",
    "#aec44b",
    "#8bf2d2",
    "#37d7a5",
    "#fcfe78",
    "#22ed90",
    "#80cd47",
    "#849fe7"
]


def color_for_agent_type(agent_type):
    """Utility function that map a color to an agent types"""
    # return color already mapped to that type

    if agent_type in _mapped_colors:
        return _mapped_colors[agent_type]

    # map a color to that new type
    if len(_remaining_colors) > 0:
        color = _remaining_colors.pop(0)
    else:
        # use default color
        color = "#aca69f"
        logging.warning(
            "not enough colors in palette for new agent type, using default one"
        )

    _mapped_colors[agent_type] = color
    return color



import logging
import os
from pathlib import Path

import fameio.input.loader as fameio
import yaml
from fameio.input.scenario.fameiofactory import FameIOFactory

import famegui.models as models
from famegui.config.config import DEFAULT_FAME_WORK_DIR


def _check_all_agents_have_display_coords(scenario: models.Scenario):
    if len(scenario.agents) == 0:
        return True
    for a in scenario.agents:
        if a.display_xy is None:
            return False
        return True


def insert_work_dir_into_paths(data, fame_work_dir):
    if isinstance(data, dict):
        for key, value in data.items():
            data[key] = insert_work_dir_into_paths(value, fame_work_dir)  # Recurse for nested dictionaries
    elif isinstance(data, list):
        data = [insert_work_dir_into_paths(item, fame_work_dir) for item in data]  # Recurse for lists
    elif isinstance(data, str) and data.startswith("./"):
        data = data.replace("./", fame_work_dir + "/", 1)  # Replace only the first instance
    return data


class CustomFameIOFactory(FameIOFactory):
    @staticmethod
    def new_schema_from_dict(definitions: dict) -> models.Schema:
        return models.Schema.from_dict(definitions)

    @staticmethod
    def new_general_properties_from_dict(definitions: dict) -> models.GeneralProperties:
        return models.GeneralProperties.from_dict(definitions)

    @staticmethod
    def new_agent_from_dict(definitions: dict) -> models.Agent:
        return models.Agent.from_dict(definitions)

    @staticmethod
    def new_contract_from_dict(definitions: dict) -> models.Contract:
        return models.Contract.from_dict(definitions)


class ScenarioLoader:

    @staticmethod
    def load_predefined_options():

        example_schema_path = os.path.join(DEFAULT_FAME_WORK_DIR, "schemas", "schema-example.yaml")
        example_props_path = os.path.join(DEFAULT_FAME_WORK_DIR, "props", "properties-example.yaml")

        with open(example_schema_path, 'r') as file:
            example_schema = yaml.safe_load(file)

        with open(example_props_path, 'r') as file:
            example_props = yaml.safe_load(file)

        return {"schema": example_schema,
                "GeneralProperties": example_props}

    @staticmethod
    def load_yaml_file(file_path: str, path_resolver: fameio.PathResolver,
                       fame_work_dir: str) -> models.Scenario:
        """Load (read and parse) a YAML scenario file"""
        file_path = os.path.abspath(file_path)

        yaml_dict = fameio.load_yaml(Path(file_path), path_resolver)

        yaml_dict = yaml_dict if yaml_dict is not None else ScenarioLoader.load_predefined_options()

        scenario = models.Scenario.from_dict(definitions=yaml_dict,
                                             factory=CustomFameIOFactory())


        # check if layout generation is necessary
        if not _check_all_agents_have_display_coords(scenario):
            logging.info("at least one Agent does not have graph layout coords (X,Y): applying layouting to all agents")
            models.layout_agents(scenario)
            for agent in scenario.agents:

                if agent.display_xy is not None:
                    continue
                agent.set_display_xy(0, 0)

        assert _check_all_agents_have_display_coords(scenario)

        return scenario

    @staticmethod
    def save_to_yaml_file(scenario: models.Scenario, file_path: str):
        """Save the given scenario to a YAML file"""
        logging.info("saving scenario to file {}".format(file_path))

        assert os.path.isabs(file_path)

        try:
            with open(file_path, "w") as f:
                export_to_dict = scenario.to_export_to_yaml_dict()

                yaml.dump(export_to_dict, f)
        except Exception as e:
            raise RuntimeError("failed to save scenario to file '{}'".format(file_path)) from e



import logging

import igraph as ig

HUGE_GRAPH_LENGTH = 60000
DEFAULT_GRAPH_LENGTH = 1000


def layout_agents(scenario):
    if len(scenario.agents) == 0:
        return

    logging.info("building graph layout for {} agents".format(len(scenario.agents)))
    g = ig.Graph(directed=True)

    # add nodes
    agent_ids = []
    for a in scenario.agents:
        g.add_vertex(str(a.id))
        agent_ids.append(a.id)

    # add connections
    for a in scenario.agents:
        for sender_id in a.inputs:
            g.add_edge(str(sender_id), str(a.id))
        for receiver_id in a.outputs:
            g.add_edge(str(a.id), str(receiver_id))

    layout = g.layout_kamada_kawai()
    assert len(layout.coords) == len(agent_ids)
    layout.fit_into(ig.BoundingBox(0, 0, DEFAULT_GRAPH_LENGTH, DEFAULT_GRAPH_LENGTH))

    for i, coord in enumerate(layout.coords):
        id = agent_ids[i]
        scenario.set_agent_display_xy(id, coord[0], coord[1])



from famegui.database.db_loader.db_manager import get_rows
from famegui.database.settings_db.init_db import init_db_session


def get_stored_shortcuts() -> list:
    with init_db_session() as session:
        res_list = get_rows(session, "shortcuts",
                            ["shortcut_name", "shortcut_key"])

    return res_list#



import re
import typing

from PySide6.QtCore import Signal, QObject
from PySide6.QtWidgets import QWidget, QVBoxLayout, QLabel, QLineEdit, QComboBox, QFrame

from famegui.config.style_config import get_default_input_style, get_default_err_input_style
from famegui.ui.block_item_wrapper import BlockItemWrapper
from famegui.ui.block_type_wrapper import BlockTypeWrapper
from famegui.ui.fame_input_panels import FileChooserWidgetPanel

import fameio.input.schema as schema

from famegui.ui.list_block import CustomListWidget
from famegui.utils import is_exactly_convertible_to_float, is_exactly_convertible_to_int
from famegui_runtime_helper.attribute_dict_formatting_helper import StringAttributeType


class MiniAttributeSpecs(QObject):

    def __init__(self, full_name, attr_type, is_mandatory,
                 options, parent=None):
        super().__init__(parent)
        self._full_name: str = full_name
        self._attr_type: str = attr_type
        self._is_mandatory: bool = is_mandatory
        self._options: list = options

    @property
    def attr_type(self):
        return self._attr_type

    @property
    def full_name(self):
        return self._full_name

    @property
    def is_mandatory(self):
        return self._is_mandatory

    def get_values(self) -> tuple[str, str, bool, typing.Union[None, list[str]]]:
        return self._full_name, self._attr_type, self._is_mandatory, self._options


class InputPanelWrapper(QFrame):
    """Wrapper for all types of input panels.
    It is used to validate input and style it accordingly to their type."""

    def __init__(
            self,
            parent=None,
            inner_widget: typing.Optional[QWidget] = None,
            signal_to_connect: Signal = None,
            validation_error_signal: Signal = None,
            attribute: MiniAttributeSpecs = None,
            attr_name: str = None,
            list_idx: int = None
    ):
        super().__init__(parent)
        self._attribute_spec: MiniAttributeSpecs = attribute
        self._inner_widget = inner_widget
        self.info_panel = QLabel("")
        self.attr_name = attr_name
        self._list_index = list_idx
        self.original_stylesheet = None
        self.validation_error_signal = validation_error_signal
        self.setContentsMargins(0, 0, 0, 0)

        if attr_name is None:
            matches = re.findall(r"(?<=\.)[^.]*$", self._attribute_spec.full_name)

            attr_short_name = matches[0]
            self.attr_name = attr_short_name

        if not self.validation_error_signal:
            assert False, "validation_error_signal is None"

        self._init_layout()
        if inner_widget:
            self._inner_layout.addWidget(inner_widget)

        self._inner_layout.setSpacing(0)

        if signal_to_connect:
            signal_to_connect.connect(self.validate_input)
        else:
            raise ValueError("signal_to_connect is None")

        self._inner_layout.addWidget(self.info_panel)

        uniq_obj_name = f"myUniqueWidget-{id(inner_widget)}"
        inner_widget.setObjectName(uniq_obj_name)
        inner_widget.setStyleSheet(
            get_default_input_style(uniq_obj_name)
        )
        self.original_stylesheet = inner_widget.styleSheet()
        self._signal_to_connect = signal_to_connect



    def get_attribute_spec(self):
        """Return attribute spec"""
        return self._attribute_spec

    def get_short_attr_name(self):
        """Return short attribute name"""
        return self.attr_name

    def get_attribute_full_name(self) -> str:
        """Return full attribute name"""
        return self._attribute_spec.full_name

    def _init_layout(self):
        self._inner_layout = QVBoxLayout()
        self.setLayout(self._inner_layout)

        self._inner_layout.setContentsMargins(0, 0, 0, 0)
        self._inner_layout.setSpacing(0)

    def set_list_index(self, list_idx: int):
        self._list_index = list_idx

    @property
    def list_index(self):
        if self._list_index == -1:
            raise ValueError("List index value err")

        return self._list_index

    def _has_more_than_two_dots(self, s: str) -> bool:
        """Helper function to get the short name of an attribute"""

        return s.count(".") > 0

    def validate_input(self):
        """validate input and style input panel accordingly to their type"""

        if self._attribute_spec.is_mandatory:
            if self.get_input_value() is None:

                if self._inner_widget:
                    if self._attribute_spec.attr_type == schema.AttributeType.BLOCK:
                        return

                    self.info_panel.setText(
                        f'<font color="red">Please fill in {self._attribute_spec.full_name}</font>')

                    if isinstance(self._inner_widget, FileChooserWidgetPanel):
                        self._inner_widget: FileChooserWidgetPanel
                        self._inner_widget.set_validation_state(False)

                    if self._inner_widget.objectName():
                        obj_name = self._inner_widget.objectName()
                        self._inner_widget.setStyleSheet(
                            get_default_err_input_style(obj_name)

                        )
                        self.validation_error_signal.emit(
                            False, self._attribute_spec.full_name
                        )

                        return False
                    uniq_obj_name = f"myUniqueWidget-{id(self._inner_widget)}"
                    self._inner_widget.setObjectName(uniq_obj_name)
                    self._inner_widget.setStyleSheet(
                        get_default_err_input_style(uniq_obj_name)

                    )
                self.validation_error_signal.emit(False, self._attribute_spec.full_name)

                return False

        self._inner_widget.setStyleSheet(self.original_stylesheet)
        self.info_panel.setText(f"")

        if isinstance(self._inner_widget, FileChooserWidgetPanel):
            self._inner_widget: FileChooserWidgetPanel
            self._inner_widget.set_validation_state(True)
        self.validation_error_signal.emit(True, self._attribute_spec.full_name)

        return True

    def convert_text_to_type(self, text: str, attr_type: str):
        """Convert text to the type of attribute to save it to the scenario"""

        if not text:
            return None

        if attr_type == StringAttributeType.INTEGER.value or attr_type == StringAttributeType.LONG.value:
            return int(text) if text else None
        elif attr_type == StringAttributeType.DOUBLE.value:
            return float(text) if text else None

        return text

    def get_input_value(self):
        """Get input values in form of a dict"""

        if isinstance(self._inner_widget, QLineEdit):
            return (
                self.convert_text_to_type(
                    self._inner_widget.text(), self._attribute_spec.attr_type
                )
                if self._inner_widget.text() != ""
                else None
            )

        elif isinstance(self._inner_widget, QComboBox):
            return (
                self._inner_widget.currentText()
                if self._inner_widget.currentText() != ""
                else None
            )

        elif isinstance(self._inner_widget, FileChooserWidgetPanel):
            plain_value = self._inner_widget.get_path()
            if len(plain_value) == 0:
                return None

            if is_exactly_convertible_to_float(plain_value):
                return float(self._inner_widget.get_path())

            if is_exactly_convertible_to_int(plain_value):
                return float(self._inner_widget.get_path())

            return (
                self._inner_widget.get_path()
                if self._inner_widget.get_path() != ""
                else None
            )

        elif isinstance(self._inner_widget, BlockTypeWrapper):
            self._inner_widget: BlockTypeWrapper

            return (
                self._inner_widget.get_data()
                if self._inner_widget.get_data() != ""
                else None
            )

        elif isinstance(self._inner_widget, BlockItemWrapper):
            self._inner_widget: BlockItemWrapper
            block_item_data = self._inner_widget.get_data()

            if isinstance(block_item_data, dict):
                return (
                    self._inner_widget.get_data()
                    if self._inner_widget.get_data() != ""
                    else None
                )
            if isinstance(block_item_data, str):
                return (
                    self.convert_text_to_type(
                        self._inner_widget.get_data(), self._attribute_spec.attr_type
                    )
                    if self._inner_widget.get_data() != ""
                    else None
                )

        elif isinstance(self._inner_widget, CustomListWidget):
            self._inner_widget: CustomListWidget

            return (
                self._inner_widget.get_data()
                if self._inner_widget.get_data() != ""
                else None
            )

        else:
            return f"Soon ... {str(type(self._inner_widget))}"



from famegui.agent_controller import AgentController


def build_new_contract_dlg(sender: AgentController, receiver: AgentController) -> str:
    """Returns HTML Dialog Skeleton to transform to PySideXML"""
    return f"""<html><head/><body>
    <p>Create new contract between:</p>
    <ul>
    <li>Sender: agent <b>{sender.display_id}</b> of type <b>{sender.type_name}</b></li>
    <li>Receiver: agent <b>{receiver.display_id}</b> of type <b>{receiver.type_name}</b></li>
    </ul>
    </body></html>
    """






from .observable import Observable
from .schema import Schema
from .agent_graph_layout import layout_agents
from .agent import Agent
from .contract import Contract
from .generalproperties import GeneralProperties
from .scenario import Scenario
from .scenario_loader import ScenarioLoader
from .protobuf_writer import write_protobuf_output



import os

from sqlalchemy import MetaData

from famegui.config.config import BASE_DIR
from famegui.database.settings_db.init_db import init_db_session, get_db_engine


def load_db_data():
    # Initialize MetaData
    metadata = MetaData()

    # Reflect all tables from the database
    metadata.reflect(bind=get_db_engine())

    # List all table names
    table_names = metadata.tables.keys()

    yaml_file_path = os.path.join(BASE_DIR, "config", 'settings_config.yaml')

    from famegui.database.db_loader.db_manager import load_yaml_to_db

    with init_db_session() as session:
        load_yaml_to_db(yaml_file_path, session)



import datetime
from functools import singledispatch

import ujson as json
import logging
import os.path
import time
import typing
from typing import List

import fameio.input.scenario as fameio
import fameio.input.schema as fameio_schema
from PySide6 import QtCore
from PySide6.QtCore import QObject
from fameio.input.scenario import Contract, Attribute
from icecream import ic

import famegui.models as models
from famegui.fame_tools.string_utils import get_last_substring
from famegui_runtime_helper.attribute_dict_formatting_helper import flatten_dict, build_simple_attr_dict
from famegui_runtime_helper.delta_dict_builder import DictDiffer
from famegui_runtime_helper.fame_session_cache import FameSessionCache
from famegui.config.config import BASE_DIR
from famegui.fame_tools.dict_tools import remove_none_values_and_empty_elements
from famegui.undo_utils.undo_constants import UndoActions


class Scenario(fameio.Scenario):
    class UndoManager(QObject):
        undo_signal = QtCore.Signal(int, str, str)  # agent_id, attr_name, new_value
        display_undo_dlg_signal = QtCore.Signal(str)  # serialized data

        def __init__(self):
            super().__init__()

    def connect_to_display_undo_dlg_signal(self, slot):
        self.undo_manager.display_undo_dlg_signal.connect(slot)

    def connect_to_undo_signal(self, slot):
        self.undo_signal.connect(slot)

    def trigger_undo(self):

        # get the last file in the cache

        cache_dir = os.path.join(BASE_DIR, "cache")

        youngest_file = None
        youngest_time = 0

        for filename in os.listdir(cache_dir):
            file_path = os.path.join(cache_dir, filename)
            if os.path.isfile(file_path):
                file_time = os.path.getmtime(file_path)
                if file_time > youngest_time:
                    youngest_time = file_time
                    youngest_file = file_path

        if youngest_file:
            time.ctime(youngest_time)

            with open(youngest_file, "r") as f:
                data = json.load(f)
                agent_id = data["agent_id"]
                attr_name = data["attr_name"]
                new_value = data["old_value"]

                self.undo_signal.emit(agent_id, attr_name, new_value)

    def compare_objects(self, obj1, obj2):

        attributes = ["delivery_interval", "product_name", "senderid", "receiverid", "expire_date", "first_delivery", ]

        changes = {}
        for attribute in dir(obj1):
            if not attribute.startswith('__'):

                if attribute not in attributes:
                    continue
                if attribute in changes:
                    continue
                value1 = getattr(obj1, attribute)
                value2 = getattr(obj2, attribute)
                if value1 != value2:
                    changes[attribute] = (value1, value2)
        return changes

    # Find changes between obj1 and obj2

    settings = {}

    def __init__(self, schema: fameio_schema.Schema, props: fameio.GeneralProperties):
        super().__init__(schema, props)

        self.undo_manager = self.UndoManager()
        self.undo_signal = self.undo_manager.undo_signal

    @property
    def contracts(self) -> List[Contract]:
        return super().contracts

    def del_all_related_contracts(self, agent_id: int) -> None:
        """Delete all contracts related to the given agent"""
        contracts_to_delete = []
        for idx, contract in enumerate(super().contracts[:]):
            if contract.sender_id == agent_id or contract.receiver_id == agent_id:
                contracts_to_delete.append(contract)
        for contract in contracts_to_delete:
            super().contracts.remove(contract)

    def set_agent_display_xy(self, agent_id: int, x: int, y: int) -> None:
        for a in self.agents:
            if a.id == agent_id:
                a.set_display_xy(x, y)
                return
        raise RuntimeError(f"Unknown agent with ID '{agent_id}'")

    def update_properties(self, props: models.GeneralProperties) -> bool:
        """Sync the scenario properties with the given ones, returns True if some changes"""
        if props.to_dict() != self.general_properties.to_dict():
            logging.info("Updating scenario general properties")
            self._general_props = props
            return True
        return False

    def to_export_to_yaml_dict(self):

        return self._to_dict()

    def _to_dict(self) -> dict:
        """Serializes the scenario content to a dict"""
        result = {
            Scenario.KEY_GENERAL: self.general_properties.to_dict(),
            Scenario.KEY_SCHEMA: self.schema.to_dict(),
        }
        if self.string_sets:
            result[Scenario.KEY_STRING_SETS] = {
                name: string_set.to_dict() for name, string_set in self.string_sets.items()
            }
        if self.agents:
            result[Scenario.KEY_AGENTS] = []
            for agent in self.agents:
                result[Scenario.KEY_AGENTS].append(agent.to_fame_gui_adjusted_dict())
        if self.contracts:
            result[Scenario.KEY_CONTRACTS] = []
            for contract in self.contracts:
                result[Scenario.KEY_CONTRACTS].append(contract.to_dict())

        return result

    def get_amount_of_related_contracts(self, agent_id) -> int:
        """Return the amount of directly connected agents to the given agent_id"""
        amount = 0
        for contract in self.contracts:
            if contract.sender_id == agent_id or contract.receiver_id == agent_id:
                amount += 1
        return amount

    def get_amount_connected_agents(self, agent_id: int) -> int:
        """Return the amount of directly connected agents to the given agent_id"""
        agent_id_set = set()

        for contract in self.contracts:
            if contract.sender_id == agent_id:
                agent_id_set.add(contract.receiver_id)
                continue
            if contract.receiver_id == agent_id:
                agent_id_set.add(contract.sender_id)

        return len(agent_id_set)

    def get_agents_by_type(self, type: str) -> typing.List[models.Agent]:

        return [agent for agent in self.agents if agent.type_name == type]

    def get_agent_by_id(self, agent_id: int) -> models.Agent:
        """Return the agent object model with the given agent_id"""
        for agent in self.agents:
            if agent.id == agent_id:
                return agent
        raise RuntimeError(f"Unknown agent with ID '{agent_id}'")

    def get_all_related_agents_ids(self, agent_id: int) -> List[int]:
        """Return a list of all agents related to the given agent_id"""
        agent_id_set = set()

        for contract in self.contracts:
            if contract.sender_id == agent_id:
                agent_id_set.add(contract.receiver_id)
                continue
            if contract.receiver_id == agent_id:
                agent_id_set.add(contract.sender_id)

        return list(agent_id_set)

    def add_contract(self, contract: models.Contract) -> None:
        # TODO: move here the logic located in MainController._create_contract_model
        super().add_contract(contract)

    @staticmethod
    def adjust_agent_attribute_keys(data_dict, agent_type):

        updated_data = {
            agent_type + "." + key: value
            for key, value in data_dict.items()
        }

        return updated_data

    def update_agent(self, updated_agent: models.Agent) -> None:
        """Update the agent with the given updated_agent."""

        for idx, agent in enumerate(super().agents[:]):
            if agent.id == updated_agent.id:

                updated_data = updated_agent.get_flat_agent_attributes_dict()

                flat_data = flatten_dict({}, updated_data)

                flat_data_old = agent.get_flat_agent_attributes_dict()

                attribute_diff = DictDiffer.diff_direct(flat_data_old, flat_data)

                changes_prefix = str(updated_agent.id) + "." + updated_agent.type_name

                attribute_diff = self.adjust_agent_attribute_keys(attribute_diff, changes_prefix)

                session_db = FameSessionCache()

                undo_item = {
                    "type": UndoActions.UNDO_AGENT_ATTRIBUTES_CHANGED.value,
                    "data": attribute_diff
                }

                session_db.add_item(undo_item)

                super().agents[idx].attributes.clear()

                for attr, value in updated_agent.attributes.items():

                    if isinstance(value, Attribute):
                        super().agents[idx].attributes[attr] = value
                        continue

                    if isinstance(value, dict):
                        value = {get_last_substring(key): value_sub for key, value_sub in value.items() if
                                 value_sub is not None}

                    super().agents[idx].attributes[attr] = Attribute(attr, value)

    def update_agent_from_dict(self, data_dict):
        non_full_name_values = {}

        attr_root_name_list = []

        agent_id, agent_type = None, None

        old_values = []

        for key in list(data_dict.keys()):
            splitted_attr_value_identifiers = key.split(".")
            agent_id = splitted_attr_value_identifiers[0]
            agent_type = splitted_attr_value_identifiers[1]
            root_attr_name = splitted_attr_value_identifiers[2]
            attr_root_name_list.append(root_attr_name)
            old_value = data_dict[key]["old_value"]
            old_values.append(old_value)

            sub_attr_key_list = splitted_attr_value_identifiers[2: (len(splitted_attr_value_identifiers))]

            values_from_store = {}

            if len(sub_attr_key_list) > 0:
                attr_name = ".".join(sub_attr_key_list)
                non_full_name_values[attr_name] = old_value
                attr_name = agent_type + "." + attr_name

                values_from_store[attr_name] = old_value

        for idx, agent in enumerate(super().agents[:]):
            if agent.id == int(agent_id):

                act = self.schema.agent_types[agent_type]

                for root_attr_name, old_value in zip(attr_root_name_list, old_values):
                    self.process_undo(act, agent, idx, non_full_name_values, old_value, root_attr_name)

    def process_undo(self, act, agent, idx, non_full_name_values, old_value, root_attr_name):
        root_attr = act.attributes.get(root_attr_name)
        if root_attr.has_nested_attributes:

            flat_full_name_labeled_values = agent.get_flat_agent_attributes_dict()

            non_full_name_values = remove_none_values_and_empty_elements(non_full_name_values)

            if root_attr.is_list:
                super().agents[idx].attributes[root_attr_name] = Attribute(root_attr_name,
                                                                           non_full_name_values[
                                                                               root_attr_name])
                return

            agent_type_name = agent.type_name

            flat_full_name_labeled_values.update(non_full_name_values)

            data_out = {}

            for attr_name, attr in act.attributes.items():
                iter_tmp_dict = build_simple_attr_dict(
                    attr, flat_full_name_labeled_values, agent_type_name)
                data_out[attr_name] = iter_tmp_dict[
                    attr_name] if attr_name in iter_tmp_dict else iter_tmp_dict

            data_out = remove_none_values_and_empty_elements(
                data_out
            )

            if not data_out:
                del super().agents[idx].attributes[root_attr_name]
                return

            data_out = data_out[root_attr_name]

            super().agents[idx].attributes[root_attr_name] = Attribute(root_attr_name, data_out)
            return
        if old_value is None:
            del super().agents[idx].attributes[root_attr_name]
            return
        super().agents[idx].attributes[root_attr_name] = Attribute(root_attr_name, old_value)
        return

    def agent_exists(self, agent_id: int) -> bool:
        """Check if an agent with the given id exists in the scenario."""
        return any([True for agent in self.agents if agent.id == agent_id])

    def update_contract(
            self, old_contract: models.Contract, updated_contract: Contract, is_in_testing_env=False
    ) -> None:
        """Update the contract with the given updated_contract."""
        """By searching the contract with the same product_name, sender_id 
        and receiver_id. to replace it with the updated_contract!!!"""

        undo_contract_change_item_data = {
            "original_contract": old_contract.to_dict(),
            "current_contract": updated_contract.to_dict(),
        }

        undo_contract_change_item = {
            "type": UndoActions.UNDO_CONTRACT_FIELDS_CHANGED.value,
            "data": undo_contract_change_item_data
        }

        if not is_in_testing_env:
            session_cache = FameSessionCache()
            session_cache.add_item(undo_contract_change_item)

        # TODO: Future Exception: if a contract has empty required attributes

        for idx, contract in enumerate(super().contracts[:]):
            if (
                    contract.product_name == old_contract.product_name
                    and contract.sender_id == old_contract.sender_id
                    and contract.receiver_id == old_contract.receiver_id
            ):
                # changes = self.compare_objects(old_contract, updated_contract)
                # keep for debugging

                super().contracts[idx] = updated_contract
                break  #

    @property
    def agents(self) -> typing.List[models.Agent]:
        """Return the list of agents in the scenario."""
        return super().agents



import logging
import os

from fameio.input.resolver import PathResolver
from fameio.input.validator import SchemaValidator
from fameio.input.writer import ProtoWriter

import famegui.models as models


def write_protobuf_output(scenario: models.Scenario, output_path: str, path_resolver: PathResolver) -> None:
    logging.debug("generating protobuf output to '{}'".format(output_path))
    assert os.path.isabs(output_path)

    timeseries_manager = SchemaValidator.validate_scenario_and_timeseries(scenario, path_resolver)

    writer = ProtoWriter(output_path, timeseries_manager)
    writer.write_validated_scenario(scenario)
    logging.info("protobuf output was generated to '{}'".format(output_path))



from PySide6 import QtWidgets
from PySide6.QtWidgets import QComboBox


def gen_preset_interval_chooser(choices: list, on_preset_changed, input_field,
                                choices_bundles, preset_choice=None):
    """
    Generates a simple ComboBox/Preset chooser Widget to choose from presets
    """

    choices.append("Custom")
    interval_preset_chooser = QtWidgets.QComboBox()
    interval_preset_chooser.addItems(choices)

    interval_preset_chooser.currentIndexChanged.connect(lambda idx: on_preset_changed(idx, choices[idx]))

    if preset_choice:
        interval_preset_chooser.setCurrentText(preset_choice)

    input_field.textChanged.connect(lambda new_text: _on_delivery_interval_steps_text_input_changed(new_text,
                                                                                                    interval_preset_chooser,
                                                                                                    choices_bundles))

    return interval_preset_chooser


def find_preset_by_seconds(presets, seconds):
    return next(
        (preset for preset in presets if preset['preset_duration_in_seconds'] == str(seconds)),
        None
    )


def _on_delivery_interval_steps_text_input_changed(text: str,
                                                   interval_preset_chooser: QComboBox,
                                                   choices_bundles: list[dict]):
    """Adjust """
    choices = []
    for i in range(interval_preset_chooser.count()):
        choices.append(interval_preset_chooser.itemText(i))

    preset_bundle = find_preset_by_seconds(choices_bundles, text)

    interval_preset_chooser.blockSignals(True)

    if preset_bundle is not None:
        interval_preset_chooser.setCurrentText(preset_bundle["preset_label"])
    else:
        interval_preset_chooser.setCurrentText("Custom")

    interval_preset_chooser.blockSignals(False)



def get_pre_defined_string_set_values_from_scenario(scenario_model):
    assert scenario_model is not None

    string_set_data_dict = {}

    for item_key, item in scenario_model.string_sets.items():
        string_set_data_dict[item_key] = [key for key in item.to_dict()["values"].keys()]

    return string_set_data_dict



from famegui.models import Observable
import unittest


class Observer:
    notified_count = 0

    def __init__(self, observable: Observable):
        observable.attach(self.notified)

    def notified(self) -> None:
        self.notified_count += 1


class TestObservable(unittest.TestCase):
    def test_subscription(self):
        event = Observable()
        event.notify()  # notifying an empty observable should have no effect

        # attach an observer
        obs1 = Observer(event)
        self.assertEqual(obs1.notified_count, 0)

        # simple notification
        event.notify()
        self.assertEqual(obs1.notified_count, 1)
        event.notify()
        self.assertEqual(obs1.notified_count, 2)

        # attach twice the same observer
        event.attach(obs1.notified)
        self.assertEqual(obs1.notified_count, 2)

        # should get double notification
        event.notify()
        self.assertEqual(obs1.notified_count, 4)

        # attach another observer
        obs2 = Observer(event)
        self.assertEqual(obs1.notified_count, 4)
        self.assertEqual(obs2.notified_count, 0)

        # notify all observers
        event.notify()
        self.assertEqual(obs1.notified_count, 6)
        self.assertEqual(obs2.notified_count, 1)

    def test_forwarding(self):
        event1 = Observable()
        event2 = Observable()
        obs = Observer(event2)

        event2.notify()
        self.assertEqual(obs.notified_count, 1)

        event1.notify()  # no effect
        self.assertEqual(obs.notified_count, 1)

        event2.forward(event1)
        event1.notify()
        self.assertEqual(obs.notified_count, 2)

        event2.notify()
        self.assertEqual(obs.notified_count, 3)

    def test_composition(self):
        class Property:
            def __init__(self):
                self.event = Observable()
                self._children = []

            def add_child(self, child: "Property") -> None:
                self._children.append(child)
                # propagate the event of the child
                self.event.forward(child.event)

        # create a tree of properties [p3->p2->p1]
        p1 = Property()
        p2 = Property()
        p3 = Property()
        p2.add_child(p1)
        p3.add_child(p2)

        # observe each property
        obs1 = Observer(p1.event)
        obs2 = Observer(p2.event)
        obs3 = Observer(p3.event)

        # test event propagation (notification forwarding)
        p1.event.notify()
        self.assertEqual(obs1.notified_count, 1)
        self.assertEqual(obs2.notified_count, 1)
        self.assertEqual(obs2.notified_count, 1)

        p2.event.notify()
        self.assertEqual(obs1.notified_count, 1)
        self.assertEqual(obs2.notified_count, 2)
        self.assertEqual(obs2.notified_count, 2)

        p3.event.notify()
        self.assertEqual(obs1.notified_count, 1)
        self.assertEqual(obs2.notified_count, 2)
        self.assertEqual(obs3.notified_count, 3)

        p2.event.notify()
        self.assertEqual(obs1.notified_count, 1)
        self.assertEqual(obs2.notified_count, 3)
        self.assertEqual(obs3.notified_count, 4)

        p1.event.notify()
        self.assertEqual(obs1.notified_count, 2)
        self.assertEqual(obs2.notified_count, 4)
        self.assertEqual(obs3.notified_count, 5)



from PySide6 import QtWidgets


def gen_quick_warning_modal(current_widget, title, description):
    """Generate a minimal warning modal"""
    QtWidgets.QMessageBox.warning(current_widget, title, description)



from typing import List

from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtCore import QTimer
from PySide6.QtWidgets import QDialog

from famegui import models
from famegui.agent_controller import AgentController
from famegui.database.prebuild_queries.interval_presets import get_delivery_interval_presets
from famegui.dialogs.dialog_newcontract import DialogNewContract
from famegui.generated.ui_dialog_newcontract import Ui_DialogNewContract
from famegui.ui.quick_modals import gen_quick_warning_modal


class DialogNewMultiContract(DialogNewContract):
    """Dialog for creating a multi contract"""
    fame_to_datetime_conversion_triggered: bool = False

    def _configure_line_edit_for_unsigned_int(self, line_edit: QtWidgets.QLineEdit):
        line_edit.setText("0")
        regex_uint64 = QtCore.QRegularExpression("\\d{1,20}")
        line_edit.setValidator(QtGui.QRegularExpressionValidator(regex_uint64))
        line_edit.textChanged.connect(self._update_ok_button_status)

    def _configure_line_edit_for_signed_int(self, line_edit: QtWidgets.QLineEdit):
        line_edit.setText("0")
        regex_uint64 = QtCore.QRegularExpression("-?\\d{1,20}")
        line_edit.setValidator(QtGui.QRegularExpressionValidator(regex_uint64))
        line_edit.textChanged.connect(self._update_ok_button_status)

    def _configure_line_edit_for_optional_signed_int(self, line_edit: QtWidgets.QLineEdit):
        line_edit.setText("")
        line_edit.setPlaceholderText(self.tr("optional"))
        regex_uint64 = QtCore.QRegularExpression("-?\\d{0,20}")
        line_edit.setValidator(QtGui.QRegularExpressionValidator(regex_uint64))

    def intersection(self, lst1, lst2):
        lst3 = [value for value in lst1 if value in lst2]
        return lst3

    def format_list_item_sender(self, display_id, agent_type_name):
        return f"<li>Sender: agent <b>{display_id}</b> of type <b>{agent_type_name}</b></li> \n"

    def format_list_item_receiver(self, display_id, agent_type_name):
        return f"<li>Receiver: agent <b>{display_id}</b> of type <b>{agent_type_name}</b></li> \n"

    def _is_not_valid_contract_relation(self) -> bool:
        return len(self._sender) > 1 and len(self._receiver) != len(self._sender)

    def __init__(
            self,
            sender: [],
            receiver: [],
            schema: models.Schema,
            parent=None,
    ):
        QDialog.__init__(self, parent)
        self._ui = Ui_DialogNewContract()
        self._ui.setupUi(self)
        self._sender: [AgentController] = sender
        self._receiver: [AgentController] = receiver
        self._schema: models.Schema = schema

        self._interval_presets = get_delivery_interval_presets()

        self.setWindowTitle(self.tr("New contract"))
        sender_items = " ".join([self.format_list_item_sender(x.display_id, x.type_name) for x in self._sender])
        receiver_items = " ".join([self.format_list_item_receiver(x.display_id, x.type_name) for x in self._receiver])

        list_inner = sender_items + receiver_items

        self._ui.labelDescr.setText(
            self.tr(
                "<html><head/><body>"
                "<p>Create new contract between multiple Agents :</p>"
                f"<ul>{list_inner}"
                "</body></html>"
            )
        )

        # fill possible products to select based on the sender schema agent type

        if self._is_not_valid_contract_relation():
            gen_quick_warning_modal(
                self,
                "Validation failure",
                "The selected agents are not corresponding a one to many or  one to one schema \n\n",

            )

            QTimer.singleShot(0, self.close)  # delay close to avoid the dialog to be shown

            return

        final_sender_product_names = self._schema.get_all_products_for_type(self._sender[0].type_name)

        self._ui.comboBoxProduct.addItems(list(set(final_sender_product_names)))

        # force the user to select a product except if only one is available
        if self._ui.comboBoxProduct.count() != 1:
            self._ui.comboBoxProduct.setCurrentIndex(-1)

        self._ui.comboBoxProduct.currentIndexChanged.connect(self._update_ok_button_status)
        self._update_ok_button_status()

        self._prep_input_line_edits()

        self._prep_date_time_fields()

        # fill possible products to select based on the sender schema agent type

        # force the user to select a product except if only one is available
        if self._ui.comboBoxProduct.count() != 1:
            self._ui.comboBoxProduct.setCurrentIndex(-1)

        # connect
        self._connect_slots()

    def _connect_slots(self):
        '''connect event listeners'''
        self._ui.comboBoxProduct.currentIndexChanged.connect(self._update_ok_button_status)
        self._update_ok_button_status()

        self._ui.lineFirstDeliveryTime.textChanged.connect(self._update_fame_time_text_areas)
        self._ui.lineExpirationTime.textChanged.connect(self._update_fame_time_text_areas)

    def _prep_date_time_fields(self):
        '''configure date time fields'''
        self._ui.lineFirstDeliveryNonFameTime.setCalendarPopup(True)
        self._ui.lineFirstDeliveryNonFameTime.setDisplayFormat(self.WIDGET_DATE_FORMAT)
        self._ui.lineFirstDeliveryNonFameTime.dateTimeChanged.connect(self._update_fame_times)

        self._ui.lineExpirationTimeNonFameTime.setCalendarPopup(True)
        self._ui.lineExpirationTimeNonFameTime.setDisplayFormat(self.WIDGET_DATE_FORMAT)
        self._ui.lineExpirationTimeNonFameTime.dateTimeChanged.connect(self._update_fame_times)

    def _prep_input_line_edits(self):
        '''accept uint64 numbers as specified in protobuf schema'''
        self._configure_line_edit_for_unsigned_int(self._ui.lineDeliveryInterval)
        self._configure_line_edit_for_signed_int(self._ui.lineFirstDeliveryTime)
        self._configure_line_edit_for_optional_signed_int(self._ui.lineExpirationTime)

    def _gen_one_to_many_contract_relationship(self, sender, expiration_time) -> List[models.Contract]:

        return [
            models.Contract(
                sender.id,
                recv.id,
                self._ui.comboBoxProduct.currentText(),
                int(self._ui.lineDeliveryInterval.text()),
                int(self._ui.lineFirstDeliveryTime.text()),
                expiration_time,
            ) for recv in self._receiver
        ]

    def _gen_one_to_one_contract_relationship(self, expiration_time) -> List[models.Contract]:

        return [
            models.Contract(
                sndr.id,
                rec.id,
                self._ui.comboBoxProduct.currentText(),
                int(self._ui.lineDeliveryInterval.text()),
                int(self._ui.lineFirstDeliveryTime.text()),
                expiration_time,
            )
            for rec, sndr in zip(self._receiver, self._sender)
        ]

    def make_new_contracts(self) -> List[models.Contract]:
        expiration_time_str = self._ui.lineExpirationTime.text()
        expiration_time = int(expiration_time_str) if expiration_time_str else None

        if len(self._sender) == 1 and len(self._receiver) > 1:
            # One to many Contract Relationship
            sender = self._sender[0]

            return self._gen_one_to_many_contract_relationship(sender, expiration_time)

        if len(self._sender) == len(self._receiver):
            # One to One Contract Relationship

            return self._gen_one_to_one_contract_relationship(expiration_time)
        # Not valid

        gen_quick_warning_modal(
            self,
            "Validation failure",
            "The selected agents are not corresponding a one to many or a one to one schema \n\n",
        )

        return None



from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy.orm import declarative_base

Base = declarative_base()


class RecentlyUsedProject(Base):
    __tablename__ = "recent_used_projects"
    id = Column(Integer, primary_key=True)
    path = Column(String())

    def __repr__(self):
        return f"RecentlyUsedProject(id={self.id!r}, name={self.path!r})"



import pytest

from famegui.models import Contract


def test_contract():
    c = Contract(
        sender_id=1,
        receiver_id=2,
        product_name="ProductName",
        delivery_interval=10,
        first_delivery_time=-20,
    )
    assert c.product_name == "ProductName"
    assert c.sender_id == 1
    assert c.display_sender_id == "#1"
    assert c.receiver_id == 2
    assert c.display_receiver_id == "#2"
    assert c.delivery_interval == 10
    assert c.first_delivery_time == -20
    assert c.expiration_time is None
    assert len(c.attributes) == 0


def test_contract_same_sender_receiver():
    # accept to build with same id for source and target
    c = Contract(
        sender_id=1,
        receiver_id=1,
        product_name="ProductName",
        delivery_interval=10,
        first_delivery_time=-20,
    )
    assert c.sender_id == c.receiver_id


def test_contract_bad_delivery_time():
    # negative delivery interval is not allowed
    with pytest.raises(ValueError):
        Contract(
            sender_id=1,
            receiver_id=1,
            product_name="ProductName",
            delivery_interval=-1,
            first_delivery_time=20,
        )



from PySide6.QtWidgets import QLineEdit, QComboBox


def set_placeholder_text(widget, text, enum_values=None):
    """Helper function to set the placeholder text of a widget."""
    if isinstance(widget, QLineEdit):
        widget: QLineEdit

        if text:
            widget.setText(str(text))

    if isinstance(widget, QComboBox):
        widget: QComboBox

        for idx, value in zip(range(widget.count()), enum_values):
            if value == text:
                widget.setCurrentIndex(idx)
                break



import os

from PySide6.QtGui import QIntValidator, QFont, QFontMetrics
from typing import Tuple, List, Callable

from PySide6 import QtWidgets, QtCore, QtGui
from PySide6.QtCore import QDateTime
from PySide6.QtWidgets import (
    QHBoxLayout,
    QDateTimeEdit,
    QComboBox,
    QLineEdit,
    QWidget,
    QLabel,
    QVBoxLayout,
    QTextEdit,
)
from icecream import ic

from famegui.appworkingdir import AppWorkingDir
from famegui.config.style_config import DEFAULT_INPUT_STYLE, FAME_CALENDAR_INPUT_STYLE, FAME_LINE_EDIT_STYLE_DEFAULT, \
    FAME_LINE_EDIT_STYLE_ERR_STATE, ERROR_TEXT_STYLE_LINE_EDIT
from famegui.time_utils import convert_fame_time_to_gui_datetime, convert_gui_datetime_to_fame_time, WIDGET_DATE_FORMAT
from famegui.ui.fame_ui_elements import QFameBoldLabel


class FileChooserWidgetPanel(QtWidgets.QWidget):
    """ "Class based panel to choose a file from the file system or to enter a float value"""

    current_value_changed = QtCore.Signal(str)

    def __init__(
            self,
            working_dir: AppWorkingDir,
            field_name: str = None,
            default_value: str = None,
            parent: QtWidgets.QWidget = None,
    ) -> None:
        super().__init__(parent)

        self._working_dir = working_dir
        self.layout = QtWidgets.QHBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(0)

        self.validation_state = True

        self.default_value = default_value

        self.field_name = field_name

        self._line_edit = QTextEdit()

        self._line_edit: QTextEdit
        self._line_edit.setStyleSheet(FAME_CALENDAR_INPUT_STYLE)

        self._line_edit.setWordWrapMode(QtGui.QTextOption.WrapAtWordBoundaryOrAnywhere)

        self._line_edit.setFixedHeight(50)
        self._line_edit.setFont(QFont("Arial", 12))

        self._line_edit.setPlaceholderText(
            self.tr(
                "Please enter file path, drop in a file or type in a single float value"
            )
        )

        uniq_widget_id = str(id(self._line_edit))

        self._line_edit.setObjectName(uniq_widget_id)

        self._line_edit.setStyleSheet(
            FAME_LINE_EDIT_STYLE_DEFAULT
        )
        self._line_edit.setContentsMargins(0, 0, 0, 0)

        self._line_edit.textChanged.connect(self._on_text_edit_changed)

        self._line_edit.setStyleSheet(DEFAULT_INPUT_STYLE)

        if self.field_name:
            label = QtWidgets.QLabel(self)
            label.setFont(QFont("Arial", 8))
            label.setText(self.field_name)
            self.layout.addWidget(label)

        if self.default_value is not None:

            if field_name.__contains__(self.field_name):
                self.default_value = str(self.default_value)
                self._line_edit.setText(self.default_value)

        self.layout.addWidget(self._line_edit)
        # button (to open dialog box)
        button = QtWidgets.QPushButton("...", self)
        button.setToolTip(self.tr("Select file..."))
        # button.setFixedWidth(button.fontMetrics().width(button.text()) + 60)
        font_metrics = QFontMetrics(button.font())
        button.setFixedWidth(font_metrics.horizontalAdvance(button.text()) + 60)

        self.layout.addWidget(button)
        # connect
        button.clicked.connect(self._on_button_clicked)

        self._drop_icon_label = QtWidgets.QLabel(self)

        self.layout.addWidget(self._drop_icon_label)
        self._drop_icon_label.hide()

        self.setAcceptDrops(True)

    def set_validation_state(self, is_valid):
        """Set the validation state of the input panel after a validation check has been performed"""

        self.validation_state = is_valid

        if is_valid:
            self._line_edit.setStyleSheet(DEFAULT_INPUT_STYLE)
            return
        self._line_edit.setStyleSheet(
            FAME_LINE_EDIT_STYLE_ERR_STATE

        )

    def get_path(self) -> str:
        """Return the path of the selected file from a file chooser widget"""

        return self._line_edit.toPlainText()

    def get_input_related_widgets(self):
        return self.layout, self._line_edit

    def _set_drop_zone_style(self, is_dragging: bool) -> None:
        # Change the styling based on whether a drag event is occurring
        if is_dragging:
            style_sheet = "FileChooserWidget { border: 2px dashed #888888; }"
            self._drop_icon_label.show()
        else:
            style_sheet = ""
            self._drop_icon_label.hide()
        self.setStyleSheet(style_sheet)

    def dragLeaveEvent(self, event) -> None:
        self._set_drop_zone_style(False)

    def dropEvent(self, event) -> None:
        file_path = event.mimeData().urls()[0].toLocalFile()  # Extract the file path from the dropped item

        if os.path.isfile(file_path):
            file_path = self._working_dir.make_relative_path(file_path)
            self._line_edit.setText(file_path)
        self._set_drop_zone_style(False)

    def dragEnterEvent(self, event) -> None:
        # Check if the dragged item is a file
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent(self, event) -> None:
        # Accept drag movement
        event.accept()

    def _on_button_clicked(self) -> None:
        open_location = self._working_dir.timeseries_dir
        if self._line_edit.toPlainText() != "":
            full_edit_path = self._working_dir.make_full_path(
                self._line_edit.toPlainText()
            )
            if os.path.isfile(full_edit_path):
                open_location = full_edit_path

        file_path, _ = QtWidgets.QFileDialog.getOpenFileName(
            self,
            self.tr("Open time series file"),
            open_location,
            self.tr("Time series (*.csv);;All files (*.*)"),
        )
        if file_path != "":
            file_path = self._working_dir.make_relative_path(file_path)
            self._line_edit.setText(file_path)

    def _on_text_edit_changed(self) -> None:
        self.current_value_changed.emit(self._line_edit.toPlainText())

    def display_error(self, has_error: bool) -> None:
        style_sheet = ERROR_TEXT_STYLE_LINE_EDIT if has_error else ""
        self._line_edit.setStyleSheet(style_sheet)


class FameCalendarInputPanelWidget(QWidget):
    """Class based panels require to call the get_input_related_widgets() method
    -> get the layout and the input widget
    The input is to access the input widget
    Return a tuple of the layout and the calender input widget
    Can be stacked easily into nested and complex layouts
    Class based panels have usually a richer functionality and some intern event based logic
    """

    WIDGET_DATE_FORMAT = "dd/MM/yyyy hh:mm:ss"

    def __init__(
            self, parent: QtWidgets.QWidget, label: str, current_fame_time: int = None
    ) -> None:
        super().__init__(parent)
        self._label = label
        self._current_fame_time = current_fame_time

        self._inner_layout = QHBoxLayout(self)
        self._inner_layout.setContentsMargins(0, 0, 0, 0)
        self._inner_layout.setSpacing(0)

        self._line_edit = QLineEdit()
        self._line_edit.setValidator(QIntValidator())

        line_edit_layout = QVBoxLayout()
        line_edit_layout.setContentsMargins(10, 10, 10, 10)
        line_edit_layout.addStretch(1)  # push labels and input fields to bottom

        line_edit_heading = QLabel("Fame Time:")  # heading label for QLineEdit
        font = QFont()
        font.setBold(True)
        line_edit_heading.setFont(font)

        line_edit_layout.addWidget(line_edit_heading)
        line_edit_layout.addWidget(self._line_edit)

        self._date_time_edit = QDateTimeEdit()
        self._date_time_edit.setCalendarPopup(True)
        self._date_time_edit.setDisplayFormat(self.WIDGET_DATE_FORMAT)
        self._date_time_edit.setDateTime(QDateTime.currentDateTime())

        date_time_edit_heading = QLabel("Date Time:")  # heading label for QDateTimeEdit
        date_time_edit_heading.setFont(font)

        date_time_edit_layout = QVBoxLayout()
        date_time_edit_layout.addStretch(1)  # push labels and input fields to bottom
        date_time_edit_layout.addWidget(date_time_edit_heading)
        date_time_edit_layout.addWidget(self._date_time_edit)
        date_time_edit_layout.setContentsMargins(10, 10, 10, 10)

        if current_fame_time:
            self._line_edit.setText(str(current_fame_time))

        self._date_time_edit.setMinimumWidth(parent.width() * 0.3)

        self._inner_layout.addWidget(QFameBoldLabel(label))
        self._inner_layout.addLayout(line_edit_layout)
        self._date_time_edit.dateTimeChanged.connect(self._update_fame_times)
        self._line_edit.textChanged.connect(self._update_fame_time_text_areas)
        self._inner_layout.addLayout(date_time_edit_layout)

    def get_input_related_widgets(self) -> Tuple[QHBoxLayout, QLineEdit]:
        """Returns a tuple of the layout and the calendar QDateTimeEdit widget"""
        return self._inner_layout, self._line_edit

    def _update_fame_times(self):
        """Inner event based logic"""
        self._line_edit.setText(
            str(convert_gui_datetime_to_fame_time(self._date_time_edit.text()))
        )

    def _update_fame_time_text_areas(self):
        """Inner event based logic"""
        fame_first_delivery_time: str = self._line_edit.text()
        if fame_first_delivery_time == "" or fame_first_delivery_time == "-":
            return
        gui_start_time = convert_fame_time_to_gui_datetime(
            int(fame_first_delivery_time)
        )
        self._date_time_edit.setDateTime(gui_start_time)

        fame_expiration_time: str = self._line_edit.text()
        if fame_expiration_time == "":
            return
        gui_expiration_time_time = convert_fame_time_to_gui_datetime(
            int(fame_expiration_time)
        )
        self._date_time_edit.setDateTime(gui_expiration_time_time)


"""Function based panels"""


def get_datetime_choose_panel(
        label: str, parent=None
) -> Tuple[QHBoxLayout, QDateTimeEdit]:
    """Returns a tuple of the layout and the calender QDateTimeEdit widget
    The widget is in the layout
    Widget ref is needed to access the input more easily
    Also opens up the opportunity
    to attach event based logic to the widget,
    like the datetime change event"""

    inner_layout = QHBoxLayout()
    if parent:
        inner_layout = QHBoxLayout(parent)

    date_time_edit: QDateTimeEdit = QDateTimeEdit()
    date_time_edit.setDateTime(QDateTime.currentDateTime())
    date_time_edit.setDisplayFormat(WIDGET_DATE_FORMAT)

    inner_layout.addWidget(QFameBoldLabel(label))
    inner_layout.addWidget(date_time_edit)
    return inner_layout, date_time_edit


def get_string_chooser_panel(
        options: List[str], label: str,
        preset_value=None,
        parent=None,
        on_input_changed: Callable = None,
) -> Tuple[QHBoxLayout, QComboBox]:
    """Classic combo box panel Returns a chooser box from a list of options"""
    inner_layout = QHBoxLayout()
    if parent:
        inner_layout = QHBoxLayout(parent)

    combo_box = QtWidgets.QComboBox()
    combo_box.addItems(options)
    combo_box.currentTextChanged.connect(on_input_changed)

    combo_box.setCurrentText(preset_value if preset_value is not None else next(iter(options.keys())))
    inner_layout.addWidget(QFameBoldLabel(label))
    inner_layout.addWidget(combo_box)
    return inner_layout, combo_box


def get_integer_input_panel(
        label: str, input_help_text=None, parent=None
) -> Tuple[QHBoxLayout, QtWidgets.QLineEdit]:
    """Integer input panel Returns an input field only allowing integers"""

    inner_layout = QHBoxLayout()

    if parent:
        inner_layout = QHBoxLayout(parent)

    if not input_help_text:
        input_help_text = "Enter an integer"

    int_edit = QtWidgets.QLineEdit()
    int_edit.setPlaceholderText(input_help_text)
    int_validator = QIntValidator()  # Integer validator
    int_edit.setValidator(int_validator)
    inner_layout.addWidget(QFameBoldLabel(label))
    inner_layout.addWidget(int_edit)

    return inner_layout, int_edit


def get_double_input_panel(
        label: str, input_help_text=None, parent=None
) -> Tuple[QHBoxLayout, QtWidgets.QLineEdit]:
    """Double input panel Returns an input field only allowing integers"""

    if not input_help_text:
        input_help_text = "Enter an Double"

    return get_integer_input_panel(label, input_help_text)


def get_string_input_panel(
        label: str, parent=None
) -> Tuple[QHBoxLayout, QtWidgets.QLineEdit]:
    """String input panel. Returns an input field for string input."""

    inner_layout = QHBoxLayout()
    if parent:
        inner_layout = QHBoxLayout(parent)

    string_edit = QtWidgets.QLineEdit()
    inner_layout.addWidget(QFameBoldLabel(label))
    inner_layout.addWidget(string_edit)

    return inner_layout, string_edit



from PySide6 import QtCore, QtWidgets
from PySide6.QtGui import Qt
from PySide6.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QLineEdit,
    QFrame,
    QHBoxLayout,
    QSizePolicy,
    QFormLayout,
)
from fameio.input.schema import AttributeSpecs, AttributeType
from icecream import ic

from famegui.config.style_config import LIST_ITEM_STYLE


class BlockItemWrapper(QFrame):
    """Wrapper for all types of input panels. It it used to carry children of the blocks,  gather and manage  their
    data."""

    def __init__(self, parent=None, spec=None, attr_name=None):
        super().__init__(parent)
        self.parent = parent
        self.input_widget = None
        self.specs: AttributeSpecs = spec
        self.attr_name = attr_name
        if self.specs is None:
            raise ValueError("spec is None")

        if self.specs.attr_type == AttributeType.TIME_SERIES:
            self.vertical_row = QVBoxLayout(self)
        else:
            self.vertical_row = QHBoxLayout(
                self
            )  # QVBoxLayout(self) -> for Time Series

        if attr_name is None:
            raise ValueError("attr_name is None")

        self.vertical_row.setContentsMargins(8, 8, 8, 5)

        self.setStyleSheet(
            LIST_ITEM_STYLE
        )

        self.vertical_row.setAlignment(Qt.AlignTop)

    def get_attr_full_name(self):
        return self.specs.full_name

    def get_short_attr_name(self):
        return self.attr_name

    def get_attr_spec(self):
        return self.specs

    def add_widget(self, widget):
        """Add inner widget to the block item to display other widgets like buttons, labels, etc."""
        self.vertical_row.addWidget(widget)

        self.vertical_row.setAlignment(widget, Qt.AlignTop)

    def add_input_widget(self, widget):
        self.input_widget = widget

        self.vertical_row.addWidget(self.input_widget)

    def get_data(self):
        """Group and gather all the data from the input widget."""
        data = {}
        widget = self.input_widget

        if widget:
            if widget.__class__.__name__ == "BlockItemWrapper":
                data[self.attr_name] = widget.get_data()

                return data

            if isinstance(widget, QLineEdit):
                data[self.attr_name] = widget.text()


                return data

            if widget.__class__.__name__ == "BlockTypeWrapper":
                data[self.attr_name] = widget.get_data()

                return data

            if isinstance(widget, QWidget):
                if hasattr(widget, "get_input_value"):
                    data[self.attr_name] = widget.get_input_value()

                    return data

                else:
                    raise Exception("Widget has no attribute 'get_input_value'")

        return data



import typing

from PySide6 import QtCore

from famegui import models, colorpalette
import fameio.input.scenario.attribute as fameio

from famegui.scenario_canvas_elements.agent_canvas_item import AgentGraphItem


class AgentController(QtCore.QObject):
    """Controller attached to a FAME Agent to sync it with the views and manage the Business logic of the agent"""

    model_was_modified = QtCore.Signal()
    model_was_moved = QtCore.Signal(float, float)

    def __init__(self, agent: models.Agent, assigned_svg_color: typing.Union[str, None] = None):
        super().__init__()
        assert agent is not None
        self.model = agent
        self._scene_item = None
        self.assigned_svg_color = assigned_svg_color

        self.tree_item = None

    def set_svg_color(self, svg_color):
        self.assigned_svg_color = svg_color

    def blockSignals(self, b) -> bool:
        if b:
            self.model_was_modified.disconnect()
            self.model_was_moved.disconnect()

        return super().blockSignals(b)

    @property
    def id(self) -> int:
        return self.model.id

    @property
    def display_id(self) -> str:
        return self.model.display_id

    @property
    def type_name(self) -> str:
        return self.model.type_name

    @property
    def attributes(self) -> typing.Dict[str, fameio.Attribute]:
        return self.model.attributes

    def set_display_xy(self, x: int, y: int):
        self.model.set_display_xy(x, y)

    @property
    def tooltip_text(self) -> str:
        text = "<font size='4'><b>{}</b></font>".format(self.model.type_name)

        text += (
            "<table border=0 cellpadding=2 style='border-collapse: collapse'><tbody>"
        )
        text += "<tr><td><b>{}</b></td><td>{}</td></tr>".format(
            "ID", self.model.display_id
        )
        for attr_name, attr_value in self.model.attributes.items():
            text += "<tr><td><b>{}</b></td><td>{}</td></tr>".format(
                attr_name, attr_value.value
            )

        text += "</tbody></table>"
        return text

    @property
    def svg_color(self) -> str:
        return self.assigned_svg_color if self.assigned_svg_color is not None \
            else colorpalette.color_for_agent_type(self.type_name)

    @property
    def x(self):
        assert self.model.display_xy is not None
        return self.model.display_xy[0]

    @property
    def y(self):
        assert self.model.display_xy is not None
        return self.model.display_xy[1]

    @property
    def scene_item(self) -> AgentGraphItem:
        assert self._scene_item is not None
        return self._scene_item

    def set_scene_item(self, item: AgentGraphItem):
        assert item.type() == AgentGraphItem.Type

        assert self._scene_item is None  # should be attached to an item only once
        self._scene_item = item
        self._scene_item.position_changed = lambda x, y: self._item_position_changed(
            x, y)

    def position_changed(self):
        """Called when the data of a model is changed"""
        self.model_was_modified.emit()

    def _item_position_changed(self, x: int, y: int):
        self.model.set_display_xy(x, y)
        self.model_was_moved.emit(x, y)



import os

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

parent_dir = os.path.dirname(BASE_DIR)

DEFAULT_FAME_WORK_DIR = os.path.join(parent_dir, "testing_resources")

CUSTOM_DELIVERY_STEPS_TIME_OPTION = "Custom"



from famegui_runtime_helper.dict_hash_helper import hash_dict
from famegui.database.prebuild_queries.consts_from_settings import get_setting_value, FameGeneralRuntimeSetting


class SingletonMeta(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class FameSessionCache(metaclass=SingletonMeta):
    def __init__(self):
        self.MAX_UNDO_STEPS = int(get_setting_value(FameGeneralRuntimeSetting.AMOUNT_OF_UNDO_STEPS, 10))
        if not hasattr(self, 'items'):
            self.items = []
        if not hasattr(self, 'redo_items'):
            self.redo_items = []

    def _is_item_in_list(self, item, my_list):
        item_hash = hash_dict(item)
        for list_item in my_list:
            if hash_dict(list_item) == item_hash:
                return True
        return False

    def add_item(self, item: dict):
        if self._is_item_in_list(item, self.items):
            return

        if not isinstance(item, dict):
            raise ValueError("Item must be a dictionary.")

        if len(self.items) >= self.MAX_UNDO_STEPS:
            self.items.pop(0)

        self.items.append(item)
        self.redo_items = []  # Clear redo history

    def get_last_item(self):
        if not self.items:
            return None

        item = self.items.pop()
        self.redo_items.append(item)
        return item

    def redo_last_item(self):
        if not self.redo_items:
            return None

        item = self.redo_items.pop()
        self.items.append(item)  # Direct append to avoid clearing redo stack
        return item

    def remove_item_by_index(self, index: int):
        if 0 <= index < len(self.items):
            return self.items.pop(index)
        raise IndexError("Index out of range.")

    def get_items(self):
        return self.items


import logging
import os
import sys
import shutil
import coloredlogs
import traceback

from PySide6 import QtCore, QtWidgets, QtUiTools
from PySide6.QtGui import Qt

from famegui.mainwindow import MainWindow
from famegui.usersettings import UserSettings
from famegui.appworkingdir import AppWorkingDir
from famegui import __version__


def _tr(text):
    return QtCore.QCoreApplication.translate("app", text)


_APP_NAME = "FAME GUI"
_APP_VERSION = __version__


def _setup_logs():
    logging_format = "%(asctime)s %(levelname)s %(message)s"
    coloredlogs.install(level="INFO", fmt=logging_format)


def _init_working_dir_location() -> AppWorkingDir:
    logging.debug("initializing working dir")

    s = UserSettings()
    if s.working_dir is None:
        QtWidgets.QMessageBox.information(
            None,
            _tr(f"{_APP_NAME} working directory"),
            _tr(
                f"The {_APP_NAME} working directory is not defined for this user account.\n\n"
                "This directory will be used as the root location for all the files required or referenced by the application.\n\n"
                "Click OK to configure the path to the FAME working directory."
            ),
        )
        dir_path = QtWidgets.QFileDialog.getExistingDirectory(
            None,
            _tr("Select {} working directory").format(_APP_NAME),
            "",
            QtWidgets.QFileDialog.ShowDirsOnly,
        )
        if len(dir_path) == 0:
            raise RuntimeError(
                "can't start the application without a valid working directory"
            )
        s.set_working_dir(dir_path)

    logging.info("working directory: {}".format(s.working_dir))
    return AppWorkingDir(s.working_dir)


def _init_schemas_dir(work_dir: AppWorkingDir):
    logging.debug("initializing schemas dir")

    schemas_dir = work_dir.schemas_dir
    if os.path.isdir(work_dir.schemas_dir):
        if len(work_dir.list_existing_schema_files()) > 0:
            return
    else:
        logging.info("creating directory {}".format(schemas_dir))
        os.mkdir(schemas_dir)

    # copy the default schema
    script_dir = os.path.dirname(os.path.abspath(__file__))
    default_schema_path = os.path.join(script_dir, "data", "schema-example.yaml")
    logging.info(
        "copying default schema file '{}' to {}".format(
            default_schema_path, schemas_dir
        )
    )
    try:
        shutil.copy(default_schema_path, schemas_dir)
    except Exception as e:
        raise RuntimeError(
            "failed to copy default schema file to {}".format(schemas_dir)
        ) from e


def _init_scenarios_dir(work_dir: AppWorkingDir):
    logging.debug("initializing scenarios dir")

    scenarios_dir = work_dir.scenarios_dir
    if not os.path.isdir(scenarios_dir):
        logging.info("creating directory {}".format(scenarios_dir))
        os.mkdir(scenarios_dir)


def _init_timeseries_dir(work_dir: AppWorkingDir):
    logging.debug("initializing timeseries dir")

    timeseries_dir = work_dir.timeseries_dir
    if not os.path.isdir(timeseries_dir):
        logging.info("creating directory {}".format(timeseries_dir))
        os.mkdir(timeseries_dir)


def _init_protobuf_dir(work_dir: AppWorkingDir):
    logging.debug("initializing protobuf dir")

    protobuf_dir = work_dir.protobuf_dir
    if not os.path.isdir(protobuf_dir):
        logging.info("creating directory {}".format(protobuf_dir))
        os.mkdir(protobuf_dir)


def _init_working_dir() -> AppWorkingDir:
    work_dir = _init_working_dir_location()
    _init_schemas_dir(work_dir)
    _init_scenarios_dir(work_dir)
    _init_timeseries_dir(work_dir)
    _init_protobuf_dir(work_dir)

    logging.debug("work dir location")
    logging.debug(work_dir.root_dir)

    return work_dir


_main_wnd = None


def _excepthook(exc_type, exc_value, exc_tb):
    """Sets up logging configuration, logging error to text file"""
    logging.basicConfig(
        filename="error_log.txt",
        level=logging.ERROR,
        format="%(asctime)s - %(levelname)s - %(message)s",
        filemode="a",
    )

    logging.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_tb))

    if isinstance(exc_value, ValueError):
        logging.error("{}".format(exc_value))
    else:
        logging.error(
            "Exception: {}\n{}".format(exc_value, "".join(traceback.format_tb(exc_tb)))
        )

    msg = _tr("Error: {}.").format(exc_value)
    # remove the extra final dot from the message, if any (error from fameio...)
    if msg.endswith(".."):
        msg = msg[:-1]

    if exc_value.__cause__:
        msg += _tr("\n\nDetails: {}").format(exc_value.__cause__)
    QtWidgets.QMessageBox.critical(_main_wnd, _tr("Error"), msg)


def _is_debugger_attached() -> bool:
    """Return True if the debugger is currently attached"""
    gettrace = getattr(sys, "gettrace", lambda: None)
    return gettrace() is not None


def run():
    _setup_logs()

    if _is_debugger_attached():
        # do not handle unexpected exceptions but let them be caught by the debugger
        logging.info(
            "running inside a debugger: disabling the global exception handler"
        )
    else:
        sys.excepthook = _excepthook

    ui_file_loader = QtUiTools.QUiLoader()
    app = QtWidgets.QApplication([])

    app.setOrganizationDomain("dlr.de")
    app.setOrganizationName("DLR")
    app.setApplicationName(_APP_NAME)
    app.setApplicationVersion(_APP_VERSION)
    app.setApplicationDisplayName(
        "{} (version {})".format(app.applicationName(), app.applicationVersion())
    )
    app.setAttribute(Qt.ApplicationAttribute.AA_UseDesktopOpenGL)

    work_dir = _init_working_dir()

    logging.debug("creating main window")
    _main_wnd = MainWindow(work_dir,ui_file_loader)
    _main_wnd.show()

    if len(sys.argv) > 1:
        file_path = sys.argv[1]
        logging.debug("processing command line argument '{}'".format(file_path))
        _main_wnd.load_scenario_file(file_path)

    logging.debug("running the application")
    sys.exit(app.exec())


if __name__ == "__main__":
    run()
    logging.info("exiting the application")



from PySide6 import QtCore
from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QWidget, QVBoxLayout, QPushButton, QScrollArea,
    QFrame, QLineEdit, QHBoxLayout, QLabel
)
from icecream import ic

from famegui.data_manager.string_set_manager import get_pre_defined_string_set_values_from_scenario
from famegui.database.db_loader.db_manager import get_string_set_values_from_db
from famegui.models import Scenario
from famegui.scenario_helper import add_string_set_value


class StringSetSelectWidget(QWidget):
    on_string_sets_saved = QtCore.Signal(dict)

    def __init__(self, scenario: Scenario):
        super().__init__()
        self.setWindowTitle("Simple Scroll Area App")

        # Main layout for the app
        self.layout = QVBoxLayout(self)

        # Scroll Area
        self.scroll_area = QScrollArea()
        self.scroll_area.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.scroll_area.setWidgetResizable(True)

        # Scroll content and layout
        self.scroll_content = QWidget()
        self.scroll_layout = QVBoxLayout(self.scroll_content)
        self.scroll_layout.setAlignment(Qt.AlignmentFlag.AlignTop)

        self.scroll_content.setLayout(self.scroll_layout)
        self.scroll_area.setWidget(self.scroll_content)

        # Add scroll area to the main layout
        self.layout.addWidget(self.scroll_area)

        self._string_set_from_db = get_string_set_values_from_db()

        # Buttons

        self.setMinimumHeight(500)
        self.setMinimumWidth(650)
        self._scenario = scenario

        string_set_data_dict = get_pre_defined_string_set_values_from_scenario(self._scenario)

        self._string_set_data_dict = string_set_data_dict

        # Connect Add button to create new blocks

        self.input_data_holder = {}

        for item_key, list_of_strings in self._string_set_data_dict.items():
            sub_input_data_holder = []

            self.input_data_holder[item_key] = sub_input_data_holder

            if item_key in self._string_set_from_db:
                list_of_strings.extend(self._string_set_from_db[item_key])

            list_of_strings = set(list(list_of_strings))

            self.add_string_set_group_block(item_key, list_of_strings, sub_input_data_holder=sub_input_data_holder)

        push = QPushButton("Save String Sets")
        push.clicked.connect(self.save_string_sets)

        self.layout.insertWidget(0, push)

    def save_string_sets(self):
        """Save String Sets to db and to Scenario"""

        plain_input_data = {
            string_set_key: [string_set_line_edit.text() for string_set_line_edit in string_set_list] for
            string_set_key, string_set_list in self.input_data_holder.items()
        }

        for string_set_group_name, string_set_value_list in plain_input_data.items():
            for item in string_set_value_list:
                add_string_set_value(self._scenario, string_set_group_name, item)

        self.on_string_sets_saved.emit(plain_input_data)

    def add_list_item(self, layout, default_value="", sub_input_data_holder=None):
        """Add a list item with a text input and a remove button."""
        item_widget = QWidget()
        item_layout = QHBoxLayout(item_widget)

        item_widget.setStyleSheet("background-color:#E0E0E0")

        # Text input
        text_input = QLineEdit(default_value)
        item_layout.addWidget(text_input)

        sub_input_data_holder.append(text_input)

        # Remove button
        remove_btn = QPushButton("Remove")
        remove_btn.clicked.connect(lambda: self.remove_list_item(layout, item_widget))
        item_layout.addWidget(remove_btn)

        # Add the item widget to the layout
        layout.addWidget(item_widget)

    def remove_list_item(self, layout, item_widget):
        """Remove a specific list item."""

        matching_group = ""
        index_to_remove = -1

        for group_name, group_value_list in self.input_data_holder.items():

            if item_widget in group_value_list:
                matching_group = group_name
                break

        if index_to_remove == -1:
            raise ValueError("Index to remove cannot be null")

        self.input_data_holder[matching_group] = [item for item in self.input_data_holder[matching_group] if
                                                  item != item_widget]

        layout.removeWidget(item_widget)
        item_widget.deleteLater()

    def remove_last_list_item(self, layout, sub_input_data_holder=None):
        """Remove the last item from the layout."""
        count = layout.count()
        if count > 0:
            item = layout.itemAt(count - 1).widget()
            if item:
                sub_input_data_holder.pop()
                item.deleteLater()

    def add_string_set_group_block(self, title=None, list_of_strings=None, sub_input_data_holder=None):
        """Add a block with expandable content to the scroll area."""
        holder = QFrame()
        holder.setStyleSheet("border: none; background-color: #F7F7F7; border-radius: 8px;")
        holder_layout = QVBoxLayout(holder)
        holder_layout.setAlignment(Qt.AlignmentFlag.AlignTop)

        default_title = title if title is not None else f"Block {self.scroll_layout.count() + 1}"

        # Header with label and chevron button
        header = QWidget()
        header_layout = QHBoxLayout(header)

        header_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)

        # Title label
        title_label = QLabel(default_title)
        title_label.setStyleSheet("font-weight: bold; font-size: 16px; color: #333333;")
        header_layout.addWidget(title_label)

        # Chevron button for expand/collapse
        chevron_button = QPushButton("➖")
        chevron_button.setStyleSheet("""
            QPushButton {
                background-color: transparent;
                font-size: 16px;
                color: #333333;
                border: none;
            }
            QPushButton:hover {
                color: #DBB957;
            }
        """)

        header_layout.addWidget(chevron_button)

        # Expandable content frame
        expandable_content = QFrame()
        expandable_content_layout = QVBoxLayout(expandable_content)
        expandable_content_layout.setAlignment(Qt.AlignmentFlag.AlignTop)

        # Horizontal panel at the top of expandable content
        horizontal_panel = QFrame()
        horizontal_panel_layout = QHBoxLayout(horizontal_panel)
        horizontal_panel.setStyleSheet("background-color: #F2F2F2; padding: 5px; border-radius: 4px;")
        horizontal_panel_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)

        # Add and Remove buttons in the horizontal panel
        add_item_btn = QPushButton("Add Item")
        add_item_btn.setStyleSheet("""
            QPushButton {
                background-color: #DBB957;
                color: white;
                border-radius: 3px;
                padding: 5px;
                font-size: 12px;
            }
            QPushButton:hover {
                background-color: #C9A24D;
            }
        """)
        remove_last_item_btn = QPushButton("Remove Last Item")
        remove_last_item_btn.setStyleSheet("""
            QPushButton {
                background-color: #DBB957;
                color: white;
                border-radius: 3px;
                padding: 5px;
                font-size: 12px;
            }
            QPushButton:hover {
                background-color: #C9A24D;
            }
        """)
        horizontal_panel_layout.addWidget(add_item_btn)
        horizontal_panel_layout.addWidget(remove_last_item_btn)
        expandable_content_layout.addWidget(horizontal_panel)

        # Scroll area for expandable content
        expandable_content_scroll_area = QScrollArea()
        expandable_content_scroll_area.setWidgetResizable(True)

        # Container widget for the scroll area
        scroll_content_widget = QWidget()
        scroll_content_widget.setStyleSheet("background:#FFFFFF; border: none;")
        scroll_content_layout = QVBoxLayout(scroll_content_widget)
        scroll_content_layout.setAlignment(Qt.AlignmentFlag.AlignTop)

        if list_of_strings:
            for string_item in list_of_strings:
                self.add_list_item(scroll_content_layout, string_item, sub_input_data_holder)

        add_item_btn.clicked.connect(lambda: self.add_list_item(scroll_content_layout, "", sub_input_data_holder))
        remove_last_item_btn.clicked.connect(
            lambda: self.remove_last_list_item(scroll_content_layout, sub_input_data_holder))

        scroll_content_widget.setLayout(scroll_content_layout)
        expandable_content_scroll_area.setWidget(scroll_content_widget)
        expandable_content_layout.addWidget(expandable_content_scroll_area)

        holder.setFixedHeight(60 + 200)

        holder_layout.addWidget(header)
        holder_layout.addWidget(expandable_content)

        def toggle_expand():
            if expandable_content.isVisible():
                expandable_content.setVisible(False)
                holder.setFixedHeight(60)
                chevron_button.setText("➕")
            else:
                expandable_content.setVisible(True)
                holder.setFixedHeight(60 + 200)
                chevron_button.setText("➖")

        chevron_button.clicked.connect(toggle_expand)
        self.scroll_layout.addWidget(holder)



import typing

from sqlalchemy import MetaData, Table
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Session

from famegui.database.db_loader.db_manager import get_row_count, bulk_insert, get_rows, get_single_row
from famegui.database.settings_db.init_db import init_db_session


def colors_for_agent_types_exit_in_db() -> bool:
    """
    Check if the db has specified colors for agent types
    """
    row_count = 0
    with init_db_session() as session:
        row_count = get_row_count(session, "agent_colors")

    return row_count > 0


def get_all_colors_for_agent_types() -> typing.Union[None, typing.List[typing.Dict]]:
    """
    Retrieve all colors related to agent all available agent types
    """
    res_list = []

    with init_db_session() as session:
        res_list = get_rows(session, "agent_colors",
                            ["agent_color_hex_code", "agent_type_name"])

    return res_list


def get_color_for_agent_type(agent_type: str) -> typing.Union[str, None]:
    """
    Retrieves hex color code for given agent type from database ( if available) else None.

    Args:
        agent_type (str): The type of agent to look up

    Returns:
            str | None: Hex color code if found, None if not found
        """
    res = None
    with init_db_session() as session:
        res = get_single_row(session, "agent_colors",
                             "agent_type_name",
                             agent_type,
                             ["agent_color_hex_code"])
    return res["agent_color_hex_code"] if res is not None else None


def bulk_insert_colors(type_to_color_dict: dict):
    """
    Bulk insert colors to populate the db with randomly picked colors for every agent type
    """
    map_fields = lambda key, color: {
        "agent_color_hex_code": color,
        "agent_type_name": key,
        "data_type": 0  # string
    }

    formatted_records = [map_fields(key, item) for key, item in type_to_color_dict.items()]

    with init_db_session() as session:
        bulk_insert(session, "agent_colors", formatted_records)



import logging
import os
import random
from copy import deepcopy

from fameio.input.resolver import PathResolver

from famegui import models
from famegui.config.config import DEFAULT_FAME_WORK_DIR
from famegui.models import Contract
from famegui.undo_utils.contract_utils import swap_or_delete_contract
from famegui.undo_utils.undo_constants import UndoActions
from famegui_runtime_helper.dict_hash_helper import hash_dict
from famegui_runtime_helper.fame_session_cache import FameSessionCache


def test_contract_undo():
    assert os.path.exists(DEFAULT_FAME_WORK_DIR), f"Path does not exist: {DEFAULT_FAME_WORK_DIR}"

    file_path = os.path.join(DEFAULT_FAME_WORK_DIR, "scenario.yaml")

    scenario_model = models.ScenarioLoader.load_yaml_file(
        file_path, PathResolver(), DEFAULT_FAME_WORK_DIR)

    assert scenario_model is not None

    original_contracts = deepcopy(scenario_model.contracts)

    rdm_int = random.randint(0, len(scenario_model.contracts) - 1)

    process_random_contract_undo(original_contracts, rdm_int, scenario_model)


def process_random_contract_undo(original_contracts, rdm_int, scenario_model):
    # test if all contracts are serializable
    for idx, item in enumerate(scenario_model.contracts):
        contract_dict = item.to_dict()

        new_contract = Contract.from_dict(contract_dict)

        assert hash_dict(new_contract.to_dict()) == hash_dict(contract_dict)
    session_db = FameSessionCache()
    random_contract_pick = scenario_model.contracts[rdm_int]
    logger = logging.getLogger()
    logger.debug("random_contract_pick")

    logger.debug(random_contract_pick.to_dict())
    rdm_contract = random_contract_pick.to_dict()
    ref_contract = deepcopy(random_contract_pick.to_dict())
    rdm_contract = deepcopy(rdm_contract)

    rdm_contract["senderid"] = 62 if rdm_contract["senderid"] != 62 else 63
    contract_data = {"old_contract_data": ref_contract,
                     "current_contract_ref": rdm_contract}

    item_idx_to_del = None
    for idx, contract_item in enumerate(scenario_model.contracts):

        if hash_dict(contract_item.to_dict()) == hash_dict(ref_contract):
            item_idx_to_del = idx
            break
    assert item_idx_to_del is not None
    edited_contract = Contract.from_dict(rdm_contract)
    assert edited_contract is not None

    amount_of_contracts_pre_deletion = len(scenario_model.contracts)
    swap_or_delete_contract(scenario_model, ref_contract, None)
    amount_of_contracts_post_deletion = len(scenario_model.contracts)

    assert amount_of_contracts_pre_deletion != amount_of_contracts_post_deletion
    for contract_item in scenario_model.contracts:
        if hash_dict(contract_item.to_dict()) == hash_dict(ref_contract):
            assert ValueError("Removed Contract cannot be in Scenario")
    contract_undo_change_action = {
        "type": UndoActions.UNDO_CONTRACT_FIELDS_CHANGED.value,
        "data": contract_data
    }

    session_db.add_item(contract_undo_change_action)
    swap_or_delete_contract(scenario_model, None, ref_contract)
    current_contract_ref_idx = None
    for idx, contract_item in enumerate(scenario_model.contracts):
        if hash_dict(contract_item.to_dict()) == hash_dict(ref_contract):
            current_contract_ref_idx = idx
            break

    assert current_contract_ref_idx is not None
    contract_idx = None
    for idx, contract_item in enumerate(scenario_model.contracts):
        if hash_dict(contract_item.to_dict()) == hash_dict(ref_contract):
            contract_idx = idx
    assert contract_idx is not None

    for original_contract in original_contracts:
        matched = False
        for contract in scenario_model.contracts:
            if hash_dict(original_contract.to_dict()) == hash_dict(contract.to_dict()):
                matched = True
                break

        assert matched == True
    last_undo_item = session_db.get_last_item()
    last_contract_undo_change_action = last_undo_item["data"]

    logger.info("Checking....")
    logger.info(last_contract_undo_change_action["old_contract_data"])

    assert hash_dict(last_contract_undo_change_action["old_contract_data"]) != hash_dict(
        last_contract_undo_change_action["current_contract_ref"])
    current_contract_ref = last_contract_undo_change_action["current_contract_ref"]
    contracts_pre_undo = deepcopy(scenario_model.contracts)

    swap_or_delete_contract(scenario_model,
                            last_contract_undo_change_action["old_contract_data"],
                            current_contract_ref)
    swap_or_delete_contract(scenario_model,
                            current_contract_ref,
                            last_contract_undo_change_action["old_contract_data"])

    contracts_pre_undo_set = set([
        hash_dict(contract.to_dict()) for contract in contracts_pre_undo
    ])

    # check if contracts are same after performing an undo
    for item in scenario_model.contracts:
        assert hash_dict(item.to_dict()) in contracts_pre_undo_set



from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd
from PySide6.QtCore import QObject
from PySide6.QtWidgets import QSizePolicy
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas


def extract_title_from_file_path(file_path: str) -> str:
    """Extract title string from given `file_path`"""
    result = Path(file_path).stem
    result = result.replace("_", " ")
    return result.title()


class TimeSeriesPlotCanvas(QObject):
    class PlotCanvas(FigureCanvas):
        def __init__(self, parent=None, width=5, height=6, dpi=80, file_path=None, title=None):
            fig, self.ax = plt.subplots(figsize=(width, height), dpi=dpi)
            super().__init__(fig)
            self.setParent(parent)
            self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
            self.updateGeometry()
            self._file_path = file_path
            self._title = title

        def plot(self, x, y):
            self.ax.plot(x, y, 'r-')
            self.ax.set_title(self._title )
            self.ax.set_xlabel('Dates')
            y_label = extract_title_from_file_path(str(self._file_path))
            self.ax.set_ylabel(y_label)
            self.ax.grid(True)

            # Rotate x-axis labels for better readability
            self.ax.tick_params(axis='x', labelrotation=45)

            # Adjust tight layout to prevent overlapping
            plt.tight_layout()

            self.draw()

    def __init__(self, parent, file_path, title, start_date=None, end_date=None):
        super().__init__()

        self.canvas = self.PlotCanvas(parent, width=4, height=4, file_path=file_path, title=title)
        self.load_data_and_plot(file_path, start_date, end_date)

    def load_data_and_plot(self, file_path, start_date=None, end_date=None):
        # Load data
        data = pd.read_csv(file_path, delimiter=';')
        data.iloc[:, 0] = pd.to_datetime(data.iloc[:, 0], format='%Y-%m-%d_%H:%M:%S')

        # Filter by date range if provided
        if start_date:
            data = data[data.iloc[:, 0] >= pd.to_datetime(start_date)]
        if end_date:
            data = data[data.iloc[:, 0] <= pd.to_datetime(end_date)]

        x = data.iloc[:, 0]
        y = data.iloc[:, 1]

        # Plot data
        self.canvas.plot(x, y)



from enum import Enum


class UndoActions(Enum):
    # agent
    UNDO_AGENT_MOVED = 1
    UNDO_AGENT_ATTRIBUTES_CHANGED = 2
    UNDO_AGENTS_REARRANGED = 3
    UNDO_AGENT_DELETED = 4
    UNDO_AGENT_CREATED = 4

    # contract
    UNDO_CONTRACT_FIELDS_CHANGED = 6
    UNDO_CONTRACT_DELETION = 7
    UNDO_CONTRACT_CREATION = 8

    # properties
    UNDO_SCENARIO_PROPERTIES = 9



import fameio.input.scenario as fameio


class GeneralProperties(fameio.GeneralProperties):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def make_default(cls) -> "GeneralProperties":
        default_values = {
            "RunId": 1,
            "Simulation": {
                "StartTime": "2011-12-31_23:58:00",
                "StopTime": "2012-12-30_23:58:00",
                "RandomSeed": 1,
            },
            "Output": {
                "Interval": 100,
                "Process": 0,
            },
        }
        return super().from_dict(default_values)

    def set_simulation_start_time(self, value: int) -> None:
        self._simulation_start_time = value

    def set_simulation_stop_time(self, value: int):
        self._simulation_stop_time = value

    def set_simulation_random_seed(self, value: int):
        self._simulation_random_seed = value

    def set_output_interval(self, value: int):
        self._output_interval = value

    def set_output_process(self, value: int):
        self._output_process = value



import random
import typing

import PySide6
from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtGui import QBrush, QColor
from PySide6.QtWidgets import QGraphicsItem

import weakref
import math
import logging

from famegui.database.prebuild_queries.consts_from_settings import get_setting_value, FameContractGraphItemSettings


class ContractGraphItem(QtWidgets.QGraphicsItem):
    """Graphical representation of a contract between two agents.
    Note: do not register listeners here, instead use -> scenario_graph_view.py"""

    Type = QGraphicsItem.UserType + 1
    _highlight = False
    _single_highlight = False

    def __init__(self, sourceNode, destNode):
        # sourceNode: AgentGraphItem (sender), destNode: AgentGraphItem (receiver)
        QGraphicsItem.__init__(self)
        self.setAcceptedMouseButtons(QtCore.Qt.NoButton)
        self._arrow_size = 10.0
        self.CURVE_DEGREE = int(get_setting_value(FameContractGraphItemSettings.CONTRACT_LINE_ANGLE,
                                              20))  # Initialize CURVE_DEGREE here
        self._draw_line = QtCore.QLineF()
        self._control_point = QtCore.QPointF()
        self.source = weakref.ref(sourceNode)
        self.dest = weakref.ref(destNode)
        self.source().addEdge(self)
        self.dest().addEdge(self)
        self.adjust()
        self._dotted_line = True  # Randomly choose if line is dotted

    def type(self):
        """override QGraphicsItem.type()"""
        return ContractGraphItem.Type

    def sourceNode(self):
        return self.source()

    def setSourceNode(self, node):
        if node is None:
            self.source = None
            return
        self.source = weakref.ref(node)
        self.adjust()

    def destNode(self):
        return self.dest()

    def setDestNode(self, node):
        if node is None:
            self.dest = None
            return
        self.dest = weakref.ref(node)
        self.adjust()

    def set_highlight_mode(self, highlight_enabled):
        if not highlight_enabled:
            self._single_highlight = False

        self._highlight = highlight_enabled
        self._dotted_line = not highlight_enabled

    def set_single_highlight_mode(self, highlight_enabled):
        self._single_highlight = highlight_enabled

    def adjust(self):
        self._draw_line = QtCore.QLineF()
        if not self.source() or not self.dest():
            return
        radius = 50
        src_center = self.mapFromItem(self.source(), radius, radius)
        dest_center = self.mapFromItem(self.dest(), radius, radius)

        distance_x = dest_center.x() - src_center.x()
        distance_y = dest_center.y() - src_center.y()
        distance = math.sqrt(distance_x ** 2 + distance_y ** 2)

        if distance <= radius * 2:
            return

        offset_ratio = radius / distance
        offset_x = distance_x * offset_ratio
        offset_y = distance_y * offset_ratio

        self.prepareGeometryChange()
        line_start = src_center + QtCore.QPointF(offset_x, offset_y)
        line_end = dest_center - QtCore.QPointF(offset_x, offset_y)
        self._draw_line = QtCore.QLineF(line_start, line_end)

        # Calculate control point for the curve using CURVE_DEGREE
        mid_point = (line_start + line_end) / 2
        angle = math.atan2(distance_y, distance_x) + math.radians(90)  # Perpendicular angle
        curve_distance = distance * math.tan(math.radians(self.CURVE_DEGREE)) / 2
        self._control_point = mid_point + QtCore.QPointF(math.cos(angle) * curve_distance,
                                                         math.sin(angle) * curve_distance)

    def is_in_highlight_mode(self):
        return self._highlight

    def boundingRect(self):
        if not self.source() or not self.dest():
            return QtCore.QRectF()

        penWidth = 1
        extra = (penWidth + self._arrow_size) / 2.0
        size = QtCore.QSizeF(
            self._draw_line.x2() - self._draw_line.x1(),
            self._draw_line.y2() - self._draw_line.y1(),
        )

        control_point_bounds = QtCore.QRectF(self._control_point - QtCore.QPointF(extra, extra),
                                             QtCore.QSizeF(extra * 2, extra * 2))

        line_bounds = QtCore.QRectF(self._draw_line.p1(), size).normalized().adjusted(-extra, -extra, extra, extra)

        return line_bounds.united(control_point_bounds)

    def paint(self, painter, option, widget):
        if not self.source() or not self.dest():
            return

        # Draw the curve
        if self._draw_line.length() == 0.0:
            return

        color = QtCore.Qt.black
        pen_stroke = 2

        if self._highlight:
            color = QtGui.QColor("#ff1100")  # Red

        if self._single_highlight:
            pen_stroke = 5

        if self._dotted_line:
            pen_style = QtCore.Qt.DotLine
        else:
            pen_style = QtCore.Qt.SolidLine

        painter.setPen(
            QtGui.QPen(
                color,
                pen_stroke,
                pen_style,
                QtCore.Qt.RoundCap,
                QtCore.Qt.RoundJoin,
            )
        )

        path = QtGui.QPainterPath()
        path.moveTo(self._draw_line.p1())
        path.quadTo(self._control_point, self._draw_line.p2())
        painter.drawPath(path)

        # Draw the arrows if there's enough room.
        angle = math.acos(self._draw_line.dx() / self._draw_line.length())
        if self._draw_line.dy() >= 0:
            angle = (2.0 * math.pi) - angle

        line_end = self._draw_line.p2()
        destArrowP1 = line_end + QtCore.QPointF(
            math.sin(angle - math.pi / 3) * self._arrow_size,
            math.cos(angle - math.pi / 3) * self._arrow_size,
        )
        destArrowP2 = line_end + QtCore.QPointF(
            math.sin(angle - math.pi + math.pi / 3) * self._arrow_size,
            math.cos(angle - math.pi + math.pi / 3) * self._arrow_size,
        )

        painter.setBrush(color)
        painter.drawPolygon(
            QtGui.QPolygonF([self._draw_line.p2(), destArrowP1, destArrowP2])
        )



import logging
import typing

import fameio.input.schema as fameio
import yaml


def _must_load_yaml_file(file_path: str) -> dict:
    """Helper function to load the content of a YAML file"""
    logging.debug("loading yaml file {}".format(file_path))
    try:
        with open(file_path) as f:
            return yaml.safe_load(f)
    except Exception as e:
        raise RuntimeError("failed to load yaml file '{}'".format(file_path)) from e


class Schema(fameio.Schema):
    """Extends fameio.Schema with the features required for the GUI"""

    def __init__(self, definitions: dict):
        super().__init__(definitions)

    def agent_supports_product(
        self, agent_type: fameio.AgentType, product_name: str
    ) -> bool:
        """Check if the given agent type supports the given product"""
        return self.agent_types[agent_type.name].products.__contains__(product_name)

    def agent_type_from_name(self, name: str) -> fameio.AgentType:
        return super().agent_types[name] if name in super().agent_types else None

    def get_all_agent_types(self) -> typing.List[str]:
        """Return all agent types"""
        return list(super().agent_types.keys())

    @classmethod
    def from_dict(cls, definitions: dict) -> "Schema":
        return super().from_dict(definitions)

    def get_all_products_for_type(self, agent_type: str) -> typing.List[str]:
        """Return all products for the given agent type"""
        return super().agent_types[agent_type].products

    @staticmethod
    def load_yaml_file(file_path: str) -> "Schema":
        """Load (read and parse) a YAML scenario file"""
        logging.info("loading schema from file {}".format(file_path))

        data_dict = _must_load_yaml_file(file_path)

        return Schema.from_dict(data_dict)



import random
import typing

import PySide6
from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtGui import QBrush, QColor
from PySide6.QtWidgets import QGraphicsItem

import weakref
import math
import logging


class AgentGraphItem(QtWidgets.QGraphicsItem, QtCore.QObject):
    """A Agent node in the graph."""

    Type = QGraphicsItem.UserType + 2

    def zValue(self) -> float:
        z_value = 100
        return z_value

    # use a lambda rather than a Qt signal because we don't inherit from QObject
    position_changed = lambda self, x, y: None

    selected_for_edit = False
    start_pos = None

    def set_bg_color(self, color: str):
        self._color = color

    def set_selected_for_edit(self, selected: bool):
        self.selected_for_edit = selected
        self.update()

    def is_selected_for_edit(self):
        return self.selected_for_edit

    def setSelected(self, selected: bool, sender: bool = True) -> None:
        self._in_receiver_mode = sender
        super().setSelected(selected)

    def mousePressEvent(self, event):
        # Capture the start position
        self.start_pos = self.scenePos()

        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        # Capture the end position
        self.end_pos = self.scenePos()
        # Emit the signal with start and end positions


        super().mouseReleaseEvent(event)

    def __init__(self, agent_id, label, color):
        QtCore.QObject.__init__(self)  # Initialize QObject

        QGraphicsItem.__init__(self)
        # public
        self.edgeList = []
        # private / read only
        self._label = "#{}".format(agent_id)
        self._color = color
        self._radius = 50
        self._agent_id = agent_id
        self._links = []
        self._highlight_border = False
        self._in_receiver_mode = True
        # customize graphics item
        self._on_agent_dragged = None
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsMovable, True)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsSelectable, True)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemSendsGeometryChanges, True)

    def set_on_agent_dragged(self, func):
        self._on_agent_dragged = func

    def type(self):
        """override QGraphicsItem.type()"""
        return AgentGraphItem.Type

    def addEdge(self, edge):
        self.edgeList.append(weakref.ref(edge))
        edge.adjust()

    def set_border_highlight(self, highlight: bool):
        self._highlight_border = highlight

    @property
    def agent_id(self) -> int:
        return self._agent_id

    def add_link(self, link):
        self._links.append(link)

    def boundingRect(self):
        return QtCore.QRectF(0, 0, self._radius * 2, self._radius * 2)

    def adjust_contracts(self):
        for contract in self.edgeList:
            contract().adjust()

    def itemChange(self, change, value):
        if change == QGraphicsItem.GraphicsItemChange.ItemPositionChange:
            for edge in self.edgeList:
                if not edge():
                    continue
                edge().adjust()
            self.position_changed(self.x(), self.y())
        return QGraphicsItem.itemChange(self, change, value)

    def paint(self, painter, option, widget=None):
        """Qt paint method"""
        # main rectangle with its border
        if option.state & QtWidgets.QStyle.StateFlag.State_Selected:
            background_color = QtGui.QColor(self._color).darker()
            border_color = QtGui.QColor("#ff009a")  # red
            border_width = 3
        else:
            background_color = QtGui.QColor(self._color)
            border_color = QtGui.QColor(0, 0, 0)
            border_width = 2
        if not self._in_receiver_mode:
            background_color = QtGui.QColor(self._color).darker()
            border_color = QtGui.QColor("#40ff00")
            border_width = 3

        if self.selected_for_edit:
            background_color = QtGui.QColor(self._color).darker()
            border_color = QtGui.QColor("#4062BB")
            border_width = 4

        item_rect = self.boundingRect()

        painter.setBrush(QtGui.QBrush(background_color))
        painter.setPen(QtGui.QPen(border_color, border_width))
        painter.drawEllipse(item_rect)

        # label
        font = QtGui.QFont("Arial", 14)
        font.setStyleStrategy(QtGui.QFont.StyleStrategy.ForceOutline)
        painter.setFont(font)
        painter.setPen(QtGui.QPen(QtGui.QColor(0, 0, 0), 1))
        painter.drawText(item_rect, QtCore.Qt.AlignmentFlag.AlignCenter, self._label)



import pytest

from famegui_runtime_helper.dict_hash_helper import hash_dict


def test_hash_dict_consistency():
    """Test that hash_dict produces consistent hashes for identical dictionaries."""
    dict1 = {"key1": "value1", "key2": "value2"}
    dict2 = {"key2": "value2", "key1": "value1"}  # Same contents, different order

    hash1 = hash_dict(dict1)
    hash2 = hash_dict(dict2)

    assert hash1 == hash2, "Hashes should be the same for identical dictionaries"

def test_hash_dict_difference():
    """Test that hash_dict produces different hashes for different dictionaries."""
    dict1 = {"key1": "value1", "key2": "value2"}
    dict3 = {"key1": "value1", "key2": "different_value"}

    hash1 = hash_dict(dict1)
    hash3 = hash_dict(dict3)

    assert hash1 != hash3, "Hashes should be different for different dictionaries"

@pytest.mark.parametrize(
    "dict_input, expected_hash",
    [
        ({"key1": "value1", "key2": "value2"}, hash_dict({"key1": "value1", "key2": "value2"})),
        ({"a": 1, "b": [1, 2, 3]}, hash_dict({"a": 1, "b": [1, 2, 3]})),
        ({}, hash_dict({})),  # Empty dictionary
    ],
)

def test_hash_dict_with_parametrize(dict_input, expected_hash):
    """Test hash_dict with multiple cases using parametrize."""
    assert hash_dict(dict_input) == expected_hash, "Hash mismatch for input dictionary"


from datetime import datetime

from fameio.time import FameTime
from fameio.time import DATE_FORMAT as FAME_DATE_FORMAT

DATE_FORMAT = "%Y-%m-%d_%H:%M:%S"
WIDGET_DATE_FORMAT = "dd/MM/yyyy hh:mm:ss"
WIDGET_DATE_FORMAT_CONSOLE = "%d/%m/%Y %H:%M:%S"


def get_min_fame_date_time_obj() -> object:
    """Get the minimum date time object which represents a 0 in  FAME time"""
    date_time_str = "01/01/2000 00:00:00"
    date_time_obj = datetime.strptime(date_time_str, "%d/%m/%Y %H:%M:%S")
    return date_time_obj


def convert_gui_datetime_to_fame_time(widget_datetime_text: str) -> int:
    """Converts a string into a FAME time integer"""
    widget_datetime_text = widget_datetime_text.replace(".", "/")
    date_time_obj = datetime.strptime(widget_datetime_text, "%d/%m/%Y %H:%M:%S")
    date_string = date_time_obj.strftime(FAME_DATE_FORMAT)
    fame_time: int = FameTime.convert_datetime_to_fame_time_step(date_string)
    return fame_time


def convert_datetime_to_fame(fame_time_string) -> int:
    """Converts a string into a FAME time integer"""
    return FameTime.convert_datetime_to_fame_time_step(
        fame_time_string.strftime(FAME_DATE_FORMAT)
    )


def convert_fame_time_to_gui_datetime(fame_time_string: int) -> object:
    """Converts a FAME time integer into a datetime object"""
    datetime_string = FameTime.convert_fame_time_step_to_datetime(fame_time_string)
    date_time_obj = datetime.strptime(datetime_string, FAME_DATE_FORMAT)
    return date_time_obj



import copy
import logging

from PySide6 import QtCore, QtWidgets
from PySide6.QtCore import Qt
from PySide6.QtGui import QFont, QKeySequence, QShortcut
from PySide6.QtWidgets import QDialog, QTreeWidget, QSizePolicy, QVBoxLayout, QWidget, QDialogButtonBox
from PySide6.QtWidgets import QTreeWidgetItem
from fameio.input.scenario import Attribute
from fameio.input.schema import AttributeSpecs

from famegui.models import Agent, Scenario
from famegui.ui.attribute_row_item import AttributeTreeItem
from famegui.ui.quick_modals import gen_quick_warning_modal
from famegui_runtime_helper.attribute_dict_formatting_helper import get_sub_attr


class EditAgentDialog(QDialog):
    """Dialog for editing agent attributes"""

    trigger_input_validation = QtCore.Signal()
    handle_agent_attr_value_change_signal = QtCore.Signal(int, str, str)
    validation_error_signal = QtCore.Signal(bool, str)

    def __init__(self, scenario, working_dir, agents: [Agent], main_ctrl, ui_loader):
        super().__init__()

        self._loader = ui_loader
        self.default_agent_flat_value_dict = None
        self.validation_list = []
        self.setWindowTitle("EditAgentDialog")

        self.attr_top_tree_items = []
        self.validation_data = []

        self._agents = agents
        self._scenario = scenario
        self._working_dir = working_dir
        self._main_ctrl = main_ctrl
        self._schema = scenario.schema

        self._button_box = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)

        self._button_box.accepted.connect(self.accept)  # Close the dialog with `Accepted` result
        self._button_box.rejected.connect(self.reject)  # C

        self.init_ui()
        for item in self._agents:
            item: Agent
            tree_item = QTreeWidgetItem(self.treeWidgetShowSelectedAgents)
            tree_item.setCheckState(0, Qt.Checked)
            tree_item.setText(1, str(item.id))
            tree_item.setText(2, item.type_name)
        self.setWindowFlags(
            self.windowFlags() | Qt.WindowType.WindowMaximizeButtonHint | Qt.WindowType.Window
        )

        self.undo_shortcut = QShortcut(QKeySequence("Ctrl+Z"), self)

        self.undo_shortcut.setContext(Qt.ShortcutContext.ApplicationShortcut)  # Qt.ApplicationShortcut context

        self.exec_()

    def init_ui(self):
        # Main vertical layout for the dialog

        # Set the main layout to the dialog

        self.setMinimumHeight(800)
        self.setMinimumWidth(950)

        self.treeWidget = QTreeWidget()
        self.treeWidgetShowSelectedAgents = QTreeWidget()

        self.treeWidget.setRootIsDecorated(False)
        self.treeWidget.setColumnCount(2)
        self.treeWidget.setHeaderLabels([self.tr("Attribute"), self.tr("Value")])
        self.treeWidget.setColumnWidth(0, 200)
        self.treeWidget.setAlternatingRowColors(True)
        self.treeWidgetShowSelectedAgents.setRootIsDecorated(False)

        self.treeWidgetShowSelectedAgents: QTreeWidget
        self.treeWidgetShowSelectedAgents.setStyleSheet(
            "QTreeWidget::item { padding: 5px; }"
        )

        self.treeWidgetShowSelectedAgents.setFixedHeight(200)
        self.treeWidgetShowSelectedAgents.setColumnCount(3)
        self.treeWidgetShowSelectedAgents.setHeaderLabels(
            [self.tr("selected"), self.tr("#id"), self.tr("agent type")]
        )
        self.treeWidgetShowSelectedAgents.setColumnWidth(0, 200)
        self.treeWidgetShowSelectedAgents.setAlternatingRowColors(True)

        self.treeWidget.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)

        self.validation_error_signal.connect(self.on_err_validation)
        self.handle_agent_attr_value_change_signal.connect(self.on_agent_attr_changed)

        font = QFont()
        font.setBold(True)
        font.setPointSize(12)

        if len(self._agents) == 1:
            preloaded_agent = self._agents[0]
            preloaded_agent: Agent
            self.default_agent_flat_value_dict = preloaded_agent.get_flat_agent_attributes_dict()

            self.init_agent_data(preloaded_agent)
        else:
            first_agent: Agent = self._agents[0]
            current_agent_type = first_agent.type_name

            sub_dict = get_sub_attr(self._scenario.schema.agent_types[current_agent_type].attributes.items(),
                                    None, current_agent_type)

            self.default_agent_flat_value_dict = sub_dict
            self._reset_attributes()

        main_layout = QVBoxLayout(self)

        top_widget = QWidget()
        top_widget_layout = QVBoxLayout()
        top_widget_layout.addWidget(self.treeWidgetShowSelectedAgents)
        top_widget.setLayout(top_widget_layout)

        bottom_widget = QWidget()

        bottom_layout = QVBoxLayout()
        bottom_layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
        bottom_layout.setStretch(0, 1)
        bottom_widget.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Maximum)
        bottom_layout.addWidget(self.treeWidget, stretch=1)

        bottom_widget.setLayout(bottom_layout)
        scroll_area = QWidget()

        scroll_area.setLayout(bottom_layout)
        bottom_area = QtWidgets.QWidget()

        bottom_scroll_area = QVBoxLayout()
        bottom_scroll_area.addWidget(scroll_area, stretch=1)
        bottom_scroll_area.addWidget(self._button_box)

        bottom_area.setLayout(bottom_scroll_area)

        # Add widgets to layout with equal stretch factors
        main_layout.addWidget(top_widget, stretch=1)
        main_layout.addWidget(bottom_area, stretch=4)

        self.treeWidget.setUniformRowHeights(False)

        self._bottom_area = bottom_area

        self.setLayout(main_layout)

    def resizeEvent(self, event):
        super().resizeEvent(event)
        current_height = self.get_tree_content_height(self.treeWidget)

        new_max_bottom_area_height = self._bottom_area.height()

        if len(self.attr_top_tree_items) == 0:
            return

        attr_tree_item = self.attr_top_tree_items[0]

        attr_tree_item: AttributeTreeItem

        attr_tree_item.set_schema_widget_height(new_max_bottom_area_height)

    def reject(self):

        self._reset_attributes()
        super().reject()

    def accept(self):
        self._on_accept()
        # super().accept()

    def get_ui(self):
        return self

    def get_tree_content_height(self, tree_widget):
        total_height = 0
        for i in range(tree_widget.topLevelItemCount()):
            item = tree_widget.topLevelItem(i)
            total_height += tree_widget.visualItemRect(item).height()
            # Add height of expanded children
            if item.isExpanded():
                for j in range(item.childCount()):
                    child = item.child(j)
                    total_height += tree_widget.visualItemRect(child).height()
        return total_height

    def _group_invalid_data(self):
        """Group invalid data into a string to be displayed in a modal"""
        start_str = "The Following attributes are not valid:\n"

        for item in self.validation_data:
            for item_key, item_value in item.items():
                if item_key != "msg":
                    continue
                if not item["is_valid"]:
                    start_str += "- {}\n".format(item["msg"])

        return start_str

    def on_agent_attr_changed(self, agent_id, new_value, attr_name):

        """Receive signal from a panel item and print the new value"""

        logging.info(f"Attribute changed to: {new_value} for attribute: {attr_name}")

    def on_err_validation(self, is_valid, attr_name):
        """Receive validation error signal from  a panel item and append to validation data to be checked later"""

        self.validation_data = [
            item for item in self.validation_data if item["msg"] != attr_name
        ]

        self.validation_data.append({"is_valid": is_valid, "msg": attr_name})

    def init_agent_data(self, agent: Agent):
        """Initialize agent default data by building the rows of  tree widget"""
        self.treeWidget.clear()

        schema_attributes = self._scenario.schema.agent_types[
            agent.type_name

        ].attributes.items()

        for attr_name, attr_spec in schema_attributes:
            default_attr = None
            attr_spec: AttributeSpecs

            if attr_spec.has_nested_attributes:
                default_attr = agent.attributes.items()

                for item_key, item_value in agent.attributes.items():
                    if item_key == attr_name:
                        default_attr = item_value
                        break

                if not isinstance(default_attr, Attribute):
                    default_attr = None

            else:
                if attr_name in agent.attributes:
                    default_attr = agent.attributes[attr_name]

            if self.validation_error_signal is None:
                raise Exception("self.validation_error_signal is None")

            flat_values = agent.get_flat_agent_attributes_dict()

            agent_type = self._schema.agent_type_from_name(agent.type_name)

            sub_dict = get_sub_attr(agent_type.attributes.items(), flat_values, agent.type_name)

            item = AttributeTreeItem(
                self.treeWidget,
                attr_name,
                attr_spec,
                self._scenario,
                self._working_dir,
                onInputChanged=self.handle_agent_attr_value_change_signal,
                validation_signal=self.trigger_input_validation,
                validation_error_signal=self.validation_error_signal,
                default_value=default_attr,
                flat_default_values=sub_dict[attr_name],
                is_single_item=len(schema_attributes) == 1
            )

            self.attr_top_tree_items.append(item)

    def _on_cancel(self):
        """Close dialog"""

        self.close()

    @staticmethod
    def _match_type_and_id(agent_type, checked_agents_ids_set: set, agent: Agent):
        """Check if agent is of given type and id is in checked_agents_ids_set"""
        return agent.type_name == agent_type and agent.id in checked_agents_ids_set

    def _save_changes(self):
        '''Save changes to scenario'''

        top_lvl_tree_agents = [
            self.treeWidgetShowSelectedAgents.topLevelItem(i)
            for i in range(self.treeWidgetShowSelectedAgents.topLevelItemCount())
        ]

        checked_agents = [item for item in top_lvl_tree_agents if item.checkState(0) == Qt.Checked]
        checked_agents_ids = set([int(item.text(1)) for item in checked_agents])

        self._scenario: Scenario

        matching_agents = [
            agent
            for agent in copy.deepcopy(self._scenario.agents)
            if self._match_type_and_id(agent.type_name, checked_agents_ids, agent)
        ]

        for agent in matching_agents:
            agent.attributes.clear()
            for item in self.attr_top_tree_items:
                item: AttributeTreeItem

                if item.attr_value is not None:
                    value = item.attr_value
                    if isinstance(value, dict):
                        value: dict
                        if not value:
                            continue

                    agent.attributes[item.attr_name] = item.attr_value

            self._scenario.update_agent(agent)

        self._main_ctrl.set_unsaved_changes(True)

        self._reset_attributes()

        self.close()

    def _on_accept(self):
        '''Apply entered changes and close dialog if all data is valid'''
        self.validation_data.clear()
        self.trigger_input_validation.emit()

        # Notify the user if there are any invalid attributes with an Alert
        for validated_item in self.validation_data:
            for item_key, item_value in validated_item.items():
                if isinstance(item_value, bool):
                    if not item_value:
                        gen_quick_warning_modal(self, "Validation error", self._group_invalid_data())

                        return

        self._save_changes()

    def _reset_attributes(self):
        """Reset UI to default state"""
        self.treeWidget.clear()
        self.attr_top_tree_items.clear()
        current_agent_type = self._agents[0].type_name

        self.attr_top_tree_items.clear()

        for idx, (attr_name, attr_spec) in enumerate(self._scenario.schema.agent_types[
                                                         current_agent_type
                                                     ].attributes.items()):
            item = AttributeTreeItem(
                self.treeWidget,
                attr_name,
                attr_spec,
                self._scenario,
                self._working_dir,
                onInputChanged=self.handle_agent_attr_value_change_signal,
                validation_signal=self.trigger_input_validation,
                validation_error_signal=self.validation_error_signal,
                flat_default_values=self.default_agent_flat_value_dict if len(self._agents) == 1 else self.default_agent_flat_value_dict[attr_name]

            )
            self.attr_top_tree_items.append(item)

    def update(self) -> None:
        """Update tree view and adjust size"""
        super().update()



# test_fame_session_cache.py

import pytest

from famegui_runtime_helper.fame_session_cache import FameSessionCache


# Helper function to create unique items
def create_item(index):
    return {"item": f"Item {index}"}


@pytest.fixture(autouse=True)
def reset_cache():
    """
    Fixture to reset the FameSessionCache singleton before each test.
    """
    cache = FameSessionCache()
    cache.items.clear()
    cache.redo_items.clear()
    yield
    cache.items.clear()
    cache.redo_items.clear()


def test_singleton_behavior():
    """
    Test that FameSessionCache follows the singleton pattern.
    """
    cache1 = FameSessionCache()
    cache2 = FameSessionCache()
    assert cache1 is cache2, "FameSessionCache instances are not the same (singleton failed)."


def test_add_item():
    """
    Test adding unique items to the cache.
    """
    cache = FameSessionCache()
    item = create_item(1)
    cache.add_item(item)

    assert len(cache.get_items()) == 1
    assert cache.get_items()[0] == item


def test_add_duplicate_item():
    """
    Test that adding a duplicate item does not increase the cache size.
    """
    cache = FameSessionCache()
    item = create_item(1)
    cache.add_item(item)
    cache.add_item(item)  # Attempt to add duplicate

    assert len(cache.get_items()) == 1, "Duplicate item was added to the cache."


def test_max_size_limit():
    """
    Test that the cache does not exceed the maximum number of undo steps.
    """
    cache = FameSessionCache()
    max_steps = cache.MAX_UNDO_STEPS

    for i in range(max_steps + 2):  # Add more items than max
        cache.add_item(create_item(i))

    assert len(cache.get_items()) == max_steps, f"Cache exceeded max size of {max_steps}."
    expected_items = [create_item(i) for i in range(2, max_steps + 2)]
    assert cache.get_items() == expected_items, "Cache does not contain the expected items after exceeding max size."


def test_undo():
    """
    Test the undo functionality.
    """
    cache = FameSessionCache()
    for i in range(5):
        cache.add_item(create_item(i))

    undone_item = cache.get_last_item()
    assert undone_item == create_item(4), "Undo did not return the correct item."
    assert len(cache.get_items()) == 4, "Item was not removed from the cache after undo."
    assert len(cache.redo_items) == 1, "Undo did not add the item to redo_items."


def test_redo():
    """
    Test the redo functionality after an undo.
    """
    cache = FameSessionCache()
    for i in range(3):
        cache.add_item(create_item(i))

    undone_item = cache.get_last_item()
    assert undone_item == create_item(2), "Undo did not return the correct item."

    redone_item = cache.redo_last_item()
    assert redone_item == create_item(2), "Redo did not return the correct item."
    assert len(cache.get_items()) == 3, "Item was not re-added to the cache after redo."
    assert len(cache.redo_items) == 0, "Redo did not remove the item from redo_items."


def test_clear_redo_stack_on_new_addition():
    """
    Ensure that the redo stack is cleared when a new item is added after an undo.
    """
    cache = FameSessionCache()
    cache.add_item(create_item(1))
    cache.add_item(create_item(2))

    # Perform undo
    undone_item = cache.get_last_item()
    assert undone_item == create_item(2), "Undo did not return the correct item."
    assert len(cache.redo_items) == 1, "Undo did not add the item to redo_items."

    # Add a new item
    cache.add_item(create_item(3))
    assert len(cache.redo_items) == 0, "Redo stack was not cleared after adding a new item."


def test_remove_item_by_index():
    """
    Test removing an item by index.
    """
    cache = FameSessionCache()
    items = [create_item(i) for i in range(3)]
    for item in items:
        cache.add_item(item)

    removed_item = cache.remove_item_by_index(1)
    assert removed_item == create_item(1), "Incorrect item was removed by index."
    assert len(cache.get_items()) == 2, "Item was not removed correctly."
    assert cache.get_items() == [create_item(0), create_item(2)], "Cache items do not match expected after removal."


def test_get_items():
    """
    Test retrieving the current list of items.
    """
    cache = FameSessionCache()
    items = [create_item(i) for i in range(5)]
    for item in items:
        cache.add_item(item)

    retrieved_items = cache.get_items()
    assert retrieved_items == items, "Retrieved items do not match the added items."


def test_get_last_item_empty_cache():
    """
    Test getting the last item from an empty cache.
    """
    cache = FameSessionCache()
    last_item = cache.get_last_item()
    assert last_item is None, "get_last_item should return None when cache is empty."


def test_redo_without_undo():
    """
    Test that redo does nothing if no undo has been performed.
    """
    cache = FameSessionCache()
    cache.add_item(create_item(1))
    redone_item = cache.redo_last_item()
    assert redone_item is None, "Redo should return None when there is nothing to redo."
    assert len(cache.get_items()) == 1, "Cache should remain unchanged when redo is not possible."


def test_undo_until_empty():
    """
    Test undoing all items until the cache is empty.
    """
    cache = FameSessionCache()
    items = [create_item(i) for i in range(3)]
    for item in items:
        cache.add_item(item)

    for i in reversed(range(3)):
        undone_item = cache.get_last_item()
        assert undone_item == items[i], f"Undo did not return the correct item {i}."

    assert len(cache.get_items()) == 0, "Cache should be empty after undoing all items."
    assert len(cache.redo_items) == 3, "All undone items should be in redo_items."


def test_redo_after_multiple_undos():
    """
    Test performing multiple undos followed by redos.
    """
    cache = FameSessionCache()
    items = [create_item(i) for i in range(5)]
    for item in items:
        cache.add_item(item)

    # Perform two undos
    undone_item1 = cache.get_last_item()
    undone_item2 = cache.get_last_item()
    assert undone_item1 == items[4], "First undo did not return the correct item."
    assert undone_item2 == items[3], "Second undo did not return the correct item."

    # Perform two redos
    redone_item1 = cache.redo_last_item()
    redone_item2 = cache.redo_last_item()
    assert redone_item1 == items[3], "First redo did not return the correct item."
    assert redone_item2 == items[4], "Second redo did not return the correct item."
    assert len(cache.get_items()) == 5, "Cache should have all items after redos."
    assert len(cache.redo_items) == 0, "Redo stack should be empty after all redos."


def test_add_non_dict_item():
    """
    Test that adding a non-dictionary item raises a ValueError.
    """
    cache = FameSessionCache()
    with pytest.raises(ValueError):
        cache.add_item(["not", "a", "dict"])



import pytest

from famegui.models import Agent
import fameio.input.scenario.attribute as fameio


def test_agent():
    a = Agent(1, "FuelsMarket")
    assert a.id == 1
    assert a.display_id == "#1"
    assert a.type_name == "FuelsMarket"
    assert len(a.inputs) == 0
    assert len(a.outputs) == 0
    assert len(a.attributes) == 0

    a.add_attribute("NATURAL_GAS", fameio.Attribute("attr_name", "prices_gas.csv"))
    assert len(a.attributes) == 1
    assert a.attributes["NATURAL_GAS"].value == "prices_gas.csv"

    # try to add with same name
    with pytest.raises(ValueError):
        a.add_attribute("NATURAL_GAS", fameio.Attribute("attr_name", "other.csv"))

    # check it didn't change
    assert a.attributes["NATURAL_GAS"].value == "prices_gas.csv"



import typing

import fameio.input.scenario as fameio
from fameio.input.scenario import Attribute, Agent

from famegui_runtime_helper.attribute_dict_formatting_helper import flatten_dict


class Agent(fameio.Agent):
    """Extends fameio.Agent with the features required for the GUI"""

    def __init__(self, agent_id: int, type_name: str):
        super().__init__(agent_id, type_name)
        self._inputs = []
        self._outputs = []
        self._display_xy = None

    @property
    def inputs(self) -> typing.List[int]:
        return self._inputs

    def add_input(self, agent_id: int) -> None:
        self._inputs.append(agent_id)

    @property
    def outputs(self) -> typing.List[int]:
        return self._outputs

    def add_output(self, agent_id: int) -> None:
        self._outputs.append(agent_id)

    def has_display_xy(self):
        return self._display_xy is not None

    @property
    def display_xy(self) -> typing.Optional[typing.List[float]]:
        return self._display_xy

    def set_display_xy(self, x: float, y: float) -> None:
        self._display_xy = [x, y]

    @classmethod
    def from_dict(cls, definitions: dict) -> Agent:

        agent_model = super().from_dict(definitions)

        x_y_coords = None

        if "ext" in definitions:
            if "DisplayXY" in definitions["ext"]:
                x_y_coords =  definitions["ext"]["DisplayXY"]

        if x_y_coords is not None:
            agent_model.set_display_xy(x_y_coords[0], x_y_coords[1])

        return agent_model

    def build_nested_sub_list(self, final_dict, item):
        res_list = []
        for sub_dict in item.nested_list:
            list_item = {}
            for sub_dict_key, sub_dict_item in sub_dict.items():
                if isinstance(sub_dict_item, Attribute):
                    if sub_dict_item.has_value:
                        list_item[sub_dict_key] = sub_dict_item.value

                    if sub_dict_item.has_nested:
                        list_item[sub_dict_key] = self.build_attributes_dict(final_dict, sub_dict_item.nested)

            res_list.append(list_item)

        return res_list

    def get_flat_agent_attributes_dict(self):
        final_dict = {}
        result = self.build_attributes_dict(final_dict, self.attributes)

        flat_dict = flatten_dict(self.attributes, result)
        return flat_dict

    def build_attributes_dict(self, final_dict: dict, obj: dict[str, Attribute], depth=0):
        """Recursively build a dictionary of an object's attributes."""
        attributes_dict = {}

        for item_key, item in obj.items():
            if isinstance(item, Attribute):
                if item.has_nested:
                    attr_dict = self.build_attributes_dict(attributes_dict, item.nested)
                    attributes_dict[item_key] = attr_dict

                if item.has_value:
                    attributes_dict[item_key] = item.value

                if item.has_nested_list:
                    attributes_dict[item_key] = self.build_nested_sub_list(final_dict, item)
            else:
                attributes_dict[item_key] = item

        return attributes_dict

    def to_simple_attr_dict(self) -> dict:
        final_dict = {}
        final_dict = self.build_attributes_dict(final_dict, self.attributes)

        return final_dict

    def to_fame_gui_adjusted_dict(self):
        result = super().to_dict()

        result["ext"] = {
            "DisplayXY": self.display_xy
        }

        return result



import os

import pathlib

from sqlalchemy import create_engine, select
from sqlalchemy.engine import Engine
from sqlalchemy.engine.mock import MockConnection
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import Session
from typing import List

from famegui.config.config import DEFAULT_FAME_WORK_DIR
from famegui.database.tables import Base, RecentlyUsedProject


def get_db_engine():
    """Return a connection to the database"""
    try:
        engine = create_engine("sqlite:///database.sqlite3", echo=True, future=True)
        engine.connect()
    except OperationalError:
        path = os.path.join("sqlite:///", "data", "database.sqlite3")
        engine = create_engine(path, echo=True, future=True)

    Base.metadata.create_all(engine)
    engine: MockConnection
    return engine


def check_if_project_was_already_opened(path: str) -> RecentlyUsedProject:
    """Check if the project was already opened"""
    stmt = select(RecentlyUsedProject).where(RecentlyUsedProject.path == path)
    with Session(get_db_engine()) as session:
        return session.scalar(stmt)


def insert_recent_project_into_db(path: str):
    """Insert the current project into the database"""
    with Session(get_db_engine()) as session:
        last_project = RecentlyUsedProject(
            path=path,
        )
        session.add_all([last_project])
        session.commit()


def get_recently_opened_projects() -> List[RecentlyUsedProject]:
    """Return a list of the recently opened projects from the database"""
    with Session(get_db_engine()) as session:
        stmt = select(RecentlyUsedProject)
        return [x for x in session.scalars(stmt)]


def manage_path_for_db(path: str):
    """Manage the path for the database"""
    if not check_if_project_was_already_opened(str(path)):
        insert_recent_project_into_db(str(path))


if __name__ == "__main__":
    # DEBUG Purpose
    current_directory_path = pathlib.Path().resolve()
    if not check_if_project_was_already_opened(str(current_directory_path)):
        insert_recent_project_into_db(str(current_directory_path))
